/*************************************************************************
|  PROJECT NAME: ARM Monitor 
|*************************************************************************
|
|  DESCRIPTION :
|  A simple command line monitor to test BDI interface function together
|  with an ARM target.
|
|*************************************************************************/

/*************************************************************************
|  INCLUDES
|*************************************************************************/

#include <windows.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "elf.h"
#include "bdiifc.h"

/*************************************************************************
|  DEFINES
|*************************************************************************/

#define BDI_OKAY                0
#define ERR_MEM_ACCESS          -1206
#define ERR_FILE_ACCESS         -1207

/*************************************************************************
|  MACROS
|*************************************************************************/

/*************************************************************************
|  TYPEDEFS
|*************************************************************************/

/*************************************************************************
|  LOCALS
|*************************************************************************/

static DWORD      nextTargetAddr;  /* used for continous memory display */
static BOOL       terminate;
static BOOL       modeJTAG;


/****************************************************************************
 ****************************************************************************

    BDI_Append___ :

    Host independent helper function to append a numeric value to a buffer. 
    The bytes will be stored in the link format order (Motorola byte order).

     INPUT  : value     value
              buffer    pointer to buffer
     OUTPUT : RETURN    pointer to next byte after the stored value

 ****************************************************************************/

static BYTE* BDI_AppendChar(char value, BYTE* buffer)
{
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendCHAR */

static BYTE* BDI_AppendByte(BYTE  value, BYTE* buffer)
{
  *buffer++ = value;
  return buffer;
} /* BDI_AppendByte */

static BYTE* BDI_AppendWord(WORD value, BYTE* buffer)
{
  *buffer++ = (BYTE)(value>>8);
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendWord */

static BYTE* BDI_AppendLong(DWORD value, BYTE* buffer)
{
  *buffer++ = (BYTE)(value>>24);
  *buffer++ = (BYTE)(value>>16);
  *buffer++ = (BYTE)(value>>8);
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendLong */


static BYTE* BDI_AppendBlock(WORD count, const BYTE* data, BYTE* buffer)
{
  while (count--) *buffer++ = *data++;
  return buffer;
} /* BDI_AppendBlock */



/****************************************************************************
 ****************************************************************************

    BDI_Extract___ :

    Host independent helper function to extract a numeric value from a buffer. 
    The value must be stored in link format (Motorola byte order).

     INPUT  : buffer    pointer to buffer
     OUTPUT : value     pointer to variable where to store the value
              RETURN    pointer to next byte after the extracted value

 ****************************************************************************/

static BYTE* BDI_ExtractChar(char* value, BYTE* buffer)
{
  *value = (char)*buffer++;
  return buffer;
} /* BDI_ExtractCHAR */

static BYTE* BDI_ExtractByte(BYTE* value, BYTE* buffer)
{
  *value = *buffer++;
  return buffer;
} /* BDI_ExtractByte */

static BYTE* BDI_ExtractWord(WORD* value, BYTE* buffer)
{
  WORD x;
  x = (WORD)*buffer++;
  x = (x<<8) + (WORD)*buffer++;
  *value = x;
  return buffer;
} /* BDI_ExtractWord */

static BYTE* BDI_ExtractLong(DWORD* value, BYTE* buffer)
{
  DWORD x;
  x = (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  *value = x;
  return buffer;
} /* BDI_ExtractLong */

static BYTE* UTIL_ExtractWordLE(WORD* value, BYTE* buffer)
{
  WORD x;
  x  = (WORD)(*buffer++);
  x += ((WORD)(*buffer++) << 8);
  *value = x;
  return buffer;
} /* UTIL_ExtractWordLE */

static BYTE* UTIL_ExtractLongLE(DWORD* value, BYTE* buffer)
{
  DWORD x;
  x  = (DWORD)(*buffer++);
  x += ((DWORD)(*buffer++) << 8);
  x += ((DWORD)(*buffer++) << 16);
  x += ((DWORD)(*buffer++) << 24);
  *value = x;
  return buffer;
} /* UTIL_ExtractLongLE */


static BYTE* BDI_ExtractBlock(WORD count, BYTE* data, BYTE* buffer)
{
  while (count--) *data++ = *buffer++;
  return buffer;
} /* BDI_ExtractBlock */


/****************************************************************************
 ****************************************************************************

    UTIL_ExtractNumber

    Extract a number (0xnnn for hex). Conversion stops if illegal character.

    INPUT  : getPtr     pointer to the next character to read
    OUTPUT : value      the extracted integer
             RETURN     pointer to the next unread character

 ****************************************************************************/

static char* ExtractHex (DWORD* value,  char* getPtr)
{
  DWORD num;
  DWORD digit;
  BYTE  c;

  while (*getPtr == ' ') getPtr++;
  num = 0;
  for (;;) {
    c = *getPtr;
    if      ((c >= '0') && (c <= '9')) digit = (DWORD)(c - '0');
    else if ((c >= 'A') && (c <= 'F')) digit = (DWORD)(c - 'A' + 10);
    else if ((c >= 'a') && (c <= 'f')) digit = (DWORD)(c - 'a' + 10);
    else break;
    num <<= 4;
    num += digit;
    getPtr++;
  } /* for */
  *value = num;
  return getPtr;
} /* ExtractHex */

static char* ExtractDecimal (DWORD* value,  char* getPtr)
{
  DWORD num;
  DWORD digit;
  BYTE  c;

  while (*getPtr == ' ') getPtr++;
  num = 0;
  for (;;) {
    c = *getPtr;
    if      ((c >= '0') && (c <= '9')) digit = (DWORD)(c - '0');
    else break;
    num *= 10;
    num += digit;
    getPtr++;
  } /* for */
  *value = num;
  return getPtr;
} /* ExtractDecimal */


char* UTIL_ExtractNumber (DWORD* value,  char* getPtr)
{
  while (*getPtr == ' ') getPtr++;
  if ((*getPtr == '0') && ((*(getPtr+1) == 'x') || (*(getPtr+1) == 'X'))) {
    getPtr +=2;
    return ExtractHex(value, getPtr);
  } /* if */
  else {
    return ExtractDecimal(value, getPtr);
  } /* else */
} /* UTIL_ExtractNumber */


/****************************************************************************
 ****************************************************************************

    UTIL_ExtractString

    Extract a string until the next seperator.

    INPUT  : getPtr       pointer to the next character to read
    OUTPUT : string       a pointer to the first character of the string
             RETURN       pointer to the next unread character after the enumaration

 ****************************************************************************/

char* UTIL_ExtractString (char* string, char* getPtr)
{

  /* skip spaces */
  while (*getPtr == ' ') getPtr++;

  /* copy string */
  while ((*getPtr != ' ') && (*getPtr != ';') && (*getPtr != 0)) {
    *string++ = *getPtr++;
  } /* while */

  *string = 0;
  return getPtr;
} /* UTIL_ExtractString */


/****************************************************************************
 ****************************************************************************

    UTIL_ExtractEOL

    Extract a string until end of line.

    INPUT  : getPtr       pointer to the next character to read
    OUTPUT : string       a pointer to the first character of the string
             RETURN       pointer to the next unread character

 ****************************************************************************/

char* UTIL_ExtractEOL (char* string, char* getPtr)
{

  /* skip spaces */
  while (*getPtr == ' ') getPtr++;

  /* copy string */
  while (*getPtr != 0) {
    *string++ = *getPtr++;
  } /* while */

  *string = 0;
  return getPtr;
} /* UTIL_ExtractEOL */


/****************************************************************************
 ****************************************************************************

    UTIL_MatchName

    Searches for a name in a name table. The search is not case sensitive.
    All transmitted characters until the next special character are compared.
    At least one name character must be present.

    INPUT  : tableSize  the number of enties in the name table
             nameTable  the name table to search
             getPtr     pointer to the command frame
    OUTPUT : index      the index if a match is found or -1 if no match
             RETURN     pointer to the next unread character in the command frame

 ****************************************************************************/

char* UTIL_MatchName(int tableSize, char** nameTable, int* index, char* getPtr)
{
  int   i;
  char  firstChar;
  char  c;
  char* name;
  char* reference;

  firstChar = *getPtr;
  if (firstChar >= 'a') firstChar -= ('a'-'A'); /* make upper */
  for (i=0; i < tableSize; i++) {
    reference = *nameTable;
    if (*reference++ == firstChar) {
      name = getPtr;
      do {
        name++;
        c = *name;

        /* valid end */
        if (c <= ' ') {
          *index = i;
          return name;
        } /* if */

        if (c >= 'a') c -= ('a'-'A'); /* make upper */

        /* check invalid char */
        if (!(c >= 'A' && c <= 'Z') && !(c >= '0' && c <= '9')) {
          *index = -1;
          return name;
        } /* if */

      } while (*reference++ == c);

    } /* if */
    nameTable++;
  } /* for */

  /* no match found */
  *index = -1;
  return getPtr;
} /* UTIL_MatchName */


/****************************************************************************
 ****************************************************************************

  TARGET_ReadGPR:
  TARGET_ReadPSR:

  Read a Register.

  INPUT  : regNbr               Number of register
  OUTPUT : value                data
           RETURN               error code

 ****************************************************************************/

static int TARGET_ReadGPR(WORD regNbr, DWORD* value)
{
  BDI_RegTypeT    reg;

  reg.regType = BDI_RT_ARM_GPR;
  reg.regAddr = regNbr;
  return BDI_RegisterGet(1, &reg, value);
} /* TARGET_ReadGPR */

static int TARGET_ReadCPSR(DWORD* value)
{
  BDI_RegTypeT    reg;

  reg.regType = BDI_RT_ARM_CPSR;
  reg.regAddr = 0;
  return BDI_RegisterGet(1, &reg, value);
} /* TARGET_ReadCPSR */


/****************************************************************************
 ****************************************************************************

  TARGET_WriteGPR:
  TARGET_WriteICEBreaker:
  TARGET_WriteCP15:

  Write to Register.

  INPUT  : regNbr       register number
           value        the value to write
  OUTPUT : RETURN       error code

 ****************************************************************************/

static int TARGET_WriteGPR(WORD regNbr, DWORD value)
{
  BDI_RegTypeT    reg;

  reg.regType = BDI_RT_ARM_GPR;
  reg.regAddr = regNbr;
  return BDI_RegisterSet(1, &reg, &value);
} /* TARGET_WriteGPR */

int TARGET_SetTargetPC(DWORD value)
{
  BDI_RegTypeT    reg;

  reg.regType = BDI_RT_ARM_GPR;
  reg.regAddr = 15;
  return BDI_RegisterSet(1, &reg, &value);
} /* TARGET_SetTargetPC */


static int TARGET_WriteICEBreaker (WORD regNbr, DWORD value)
{
  BDI_RegTypeT    reg;

  reg.regType = BDI_RT_ARM_ICEB;
  reg.regAddr = regNbr;
  return BDI_RegisterSet(1, &reg, &value);
} /* TARGET_WriteICEBreaker */

static int TARGET_WriteCP15(WORD regNbr, DWORD value)
{
  BDI_RegTypeT    reg;

  reg.regType = BDI_RT_ARM_CP15;
  reg.regAddr = regNbr;
  return BDI_RegisterSet(1, &reg, &value);
} /* TARGET_WriteCP15 */


/****************************************************************************
 ****************************************************************************

  TARGET_ReadByte:
  TARGET_ReadWord:
  TARGET_ReadLong:

  Read from a place in target memory.
  In monitor mode, save exit from BDM is used in FREEZE mode.

  INPUT  : addr         Address of operand
  OUTPUT : value        data
           RETURN       error code

 ****************************************************************************/

int TARGET_ReadByte(DWORD addr, BYTE* value)
{
  int result;
  
  result = BDI_MemoryRead(addr, BDI_AT_DEFAULT, 1, value);
  return result;
}  /* TARGET_ReadByte */

int TARGET_ReadWord(DWORD addr, WORD* value)
{
  int   result;
  BYTE  mem[2];
  
  result = BDI_MemoryRead(addr, BDI_AT_DEFAULT, 2, mem);
  *value = (mem[1]<<8) + mem[0];
  return result;
}  /* TARGET_ReadWord */

int TARGET_ReadLong(DWORD addr, DWORD* value)
{
  int   result;
  BYTE  mem[4];
  
  result = BDI_MemoryRead(addr, BDI_AT_DEFAULT, 4, mem);
  *value = ((DWORD)mem[3]<<24) + ((DWORD)mem[2]<<16) + ((DWORD)mem[1]<<8) + (DWORD)mem[0];
  return result;
}  /* TARGET_ReadLong */


/****************************************************************************
 ****************************************************************************

  TARGET_WriteByte:
  TARGET_WriteWord:
  TARGET_WriteLong:

  Write to a place in target memory

  INPUT  : addr         Address of operand
           data         Value to write
  OUTPUT : RETURN       error code
  INOUT  : -

 ****************************************************************************/

int TARGET_WriteByte(DWORD addr, BYTE data)
{
  int result;
  
  result = BDI_MemoryWrite(addr, BDI_AT_DEFAULT, 1, &data);
  return result;
} /* TARGET_WriteByte */

int TARGET_WriteWord(DWORD addr, WORD data)
{
  int   result;
  BYTE  mem[2];
  
  mem[1] = (BYTE)(data>>8);
  mem[0] = (BYTE)(data>>0);
  result = BDI_MemoryWrite(addr, BDI_AT_DEFAULT, 2, mem);
  return result;
} /* TARGET_WriteWord */

int TARGET_WriteLong(DWORD addr, DWORD data)
{
  int   result;
  BYTE  mem[4];
  
  mem[3] = (BYTE)(data>>24);
  mem[2] = (BYTE)(data>>16);
  mem[1] = (BYTE)(data>>8);
  mem[0] = (BYTE)(data>>0);
  result = BDI_MemoryWrite(addr, BDI_AT_DEFAULT, 4, mem);
  return result;
} /* TARGET_WriteLong */


/****************************************************************************
 ****************************************************************************

    ValidEndOfLine:

    Test for a valid enf of line

    INPUT  : linePtr    pointer to the next character to read
    OUTPUT : RETURN     TRUE if valid end of line

 ****************************************************************************/

static BOOL ValidEndOfLine (const char* linePtr)
{
  while (*linePtr == ' ') linePtr++;
  return (*linePtr == 0);
} /* ValidEndOfLine */


/****************************************************************************
 ****************************************************************************

    LineFeeds :

    Sends the requested number of line feeds

    INPUT  : lineFeeds  the number of line feeds to send
    OUTPUT : -

 ****************************************************************************/

static void LineFeeds(WORD lineFeeds)
{
  while (lineFeeds--) {
    printf("\n");
  } /* while */
} /* LineFeeds */



/****************************************************************************
 ****************************************************************************

    WriteLine :

    Sends a line to the Telnet client

    INPUT  : text       the text string
    OUTPUT : -

 ****************************************************************************/

static void WriteLine(const char* text)
{
  printf("%s\n", text);
} /* WriteLine */


/****************************************************************************
 ****************************************************************************

    ReadLine :

    Reads a line from the Telnet client. Options will be directly
    processed.

    INPUT  : maxLen     the maximal length of the string
    OUTPUT : line       the read line as zero terminated string
             RETURN     the line length without trailing zero
                        a value of 0 indicates an empty line
                        a negative value indicates a closed connection

 ****************************************************************************/

static int ReadLine(WORD maxLen, char* line)
{
  gets(line);
  return (int)strlen(line);
} /* ReadLine */


/****************************************************************************
 ****************************************************************************

    DisplayError :

    Display an error line based on the error number.

    INPUT  : errorCode  the error code
             param      parameter
    OUTPUT : -

 ****************************************************************************/

void DisplayError(int errorCode, DWORD param)
{
  WriteLine(BDI_ErrorTextGet(errorCode));
} /* DisplayError */


/****************************************************************************
 ****************************************************************************

    DisplayTargetInfo :

    Display the target information.

    INPUT  : -
    OUTPUT : -

 ****************************************************************************/

/* this table must match with the data type ARM_DebugCauseT */
static char* debugCauseText[] = {
"breakpoint",
"watchpoint",
"halt forced by BDI",
"single step",
"exeption",
"semihosting"
};

static int DisplayTargetInfo(void)
{
  int       result;
  char      line[80];
  WORD      state;
  DWORD     timeStamp;
  WORD      infoCount;
  DWORD     infoValue[10];
  int       errorCode;

  infoCount = 10;
  result = BDI_StateGet(&state, &errorCode, &timeStamp, &infoCount, infoValue);
  if (result != BDI_OKAY) return result;
  switch (state) {

    case 0:
      WriteLine("    BDI state           : reset");
      break;

    case 1:
      WriteLine      ("    Target state        : debug mode");
      sprintf(line, "    Debug entry cause   : %s", debugCauseText[infoValue[0]]);
      WriteLine(line);
      sprintf(line, "    Current PC    (r15) : 0x%08lx", infoValue[1]);
      WriteLine(line);
      sprintf(line, "    Current PSR         : 0x%08lx", infoValue[2]);
      WriteLine(line);
      sprintf(line, "    Link Register (r14) : 0x%08lx", infoValue[3]);
      WriteLine(line);
      sprintf(line, "    Stackpointer  (r13) : 0x%08lx", infoValue[4]);
      WriteLine(line);
      sprintf(line, "    Vector number       : 0x%08lx", infoValue[5]);
      WriteLine(line);
      break;

    case 2:
      WriteLine("    BDI state           : running");
      break;

    case 3:
      WriteLine("    BDI state           : error");
      (void)sprintf(line, "    Error info          : %s", BDI_ErrorTextGet(errorCode));
      WriteLine(line);
      break;

    default:
      WriteLine("    BDI state         : unknown");
      break;

  } /* switch */
  
  sprintf(line, "    Reset Timestamp     : %lu", timeStamp);
  WriteLine(line);
  return BDI_OKAY;
} /* DisplayTargetInfo */


/****************************************************************************
 ****************************************************************************

    DisplayHelp :

    Display the help screen.

    INPUT  : -
    OUTPUT : -

 ****************************************************************************/

static char* helpScreen[] = {
"MD    [<address>] [<count>]  display target memory as word (32bit)",
"MDH   [<address>] [<count>]  display target memory as half word (16bit)",
"MDB   [<address>] [<count>]  display target memory as byte (8bit)",
"MM    <address>   <value>    modify word (32bit) to target memory",
"MMH   <address>   <value>    modify half word (16bit) to target memory",
"MMB   <address>   <value>    modify byte (8bit) to target memory",
"RD                           display general purpose register",
"RDI   <number>               display ICEBreaker registers",
"RDCP                         display CP15 register",
"RM    <number>   <value>     modify general purpose register",
"RMI   <number>   <value>     modify ICEBreaker register",
"RMCP  <number>   <value>     modify CP15 register",
"RESET                        reset the target system",
"STARTUP <time> <clock>       startup target with init list",
"GO    [<pc>]                 set PC and start target system",
"TI    [<pc>]                 trace on instuction (single step)",
"HALT                         force target to enter debug mode",
"INFO                         display information about the current state",
"DUMP   <addr> <count> <file> dump target memory into a binary file",
"HELP                         display command list",
"EXIT                         terminate"
};

static void DisplayHelp(void)
{
  int   i;
  char  line[80];

  for (i=0; i< (sizeof helpScreen / sizeof helpScreen[0]); i++) {
    WriteLine(helpScreen[i]);
    if ((i % 20) == 19) {
      printf("press enter to continue...");
      gets(line);
    } /* if */
  } /* for */
} /* DisplayHelp */


/****************************************************************************
 ****************************************************************************

    DisplayLong :

    Display memory words (32bit).

    INPUT  : address    the start address
             count      the number of word to display
    OUTPUT : RETURN     error code

 ****************************************************************************/

static int DisplayLong(DWORD address, WORD count)
{
  int   i;
  int   j;
  int   result;
  char  line[80];
  DWORD mem[4];

  /* display memory as block */
  result = BDI_OKAY;
  if (count == 0) {
    for (i=0; i<16; i++) {
      for (j=0; j<4; j++) {
        if (result == BDI_OKAY) result = BDI_GetLong(address + 4 * j,  &mem[j]);
      } /* for */
      if (result == BDI_OKAY) {
        (void)sprintf(line, "%08lx : %08lx %08lx %08lx %08lx", address, mem[0], mem[1], mem[2], mem[3]);
        WriteLine(line);
        address += 16;
      } /* if */
      else {
        break;
      } /* else */
    } /* for */
  } /* if */

  /* display exact the requested values */
  while ((result == BDI_OKAY && count--)) {
    result = BDI_GetLong(address, &mem[0]);
    if (result == BDI_OKAY) {
      (void)sprintf(line, "%08lx : 0x%08lx  % 11li", address, mem[0], mem[0]);
      WriteLine(line);
      address += 4;
    } /* if */
  } /* while */

  nextTargetAddr = address;
  return result;
} /* DisplayLong */


/****************************************************************************
 ****************************************************************************

    DisplayWord :

    Display memory half words (16bit).

    INPUT  : address    the start address
             count      the number of word to display
    OUTPUT : RETURN     error code

 ****************************************************************************/

static int DisplayWord(DWORD address, WORD count)
{
  int   i;
  int   j;
  int   result;
  char  line[80];
  WORD  mem[8];

  /* display memory as block */
  result = BDI_OKAY;
  if (count == 0) {
    for (i=0; i<16; i++) {
      for (j=0; j<8; j++) {
        if (result == BDI_OKAY) result = TARGET_ReadWord(address + 2 * j,  &mem[j]);
      } /* for */
      if (result == BDI_OKAY) {
        (void)sprintf(line, "%08lx : %04x %04x %04x %04x %04x %04x %04x %04x",
                              address, mem[0],mem[1],mem[2],mem[3],mem[4],mem[5],mem[6],mem[7]);
        WriteLine(line);
        address += 16;
      } /* if */
      else {
        break;
      } /* else */
    } /* for */
  } /* if */

  /* display exact the requuested values */
  while ((result == BDI_OKAY && count--)) {
    result = TARGET_ReadWord(address, &mem[0]);
    if (result == BDI_OKAY) {
      (void)sprintf(line, "%08lx : 0x%04x  % 6i", address, mem[0], mem[0]);
      WriteLine(line);
      address += 2;
    } /* if */
  } /* while */

  nextTargetAddr = address;
  return result;
} /* DisplayWord */


/****************************************************************************
 ****************************************************************************

    DisplayByte :

    Display memory as bytes (8bit).

    INPUT  : address    the start address
             count      the number of word to display
    OUTPUT : RETURN     error code

 ****************************************************************************/

static int DisplayByte(DWORD address, WORD count)
{
  int   i;
  int   j;
  int   result;
  char  line[80];
  BYTE  mem[8];

  /* display memory as block */
  result = BDI_OKAY;
  if (count == 0) {
    for (i=0; i<16; i++) {
      for (j=0; j<8; j++) {
        if (result == BDI_OKAY) result = TARGET_ReadByte(address + j,  &mem[j]);
      } /* for */
      if (result == BDI_OKAY) {
        (void)sprintf(line, "%08lx : %02x %02x %02x %02x %02x %02x %02x %02x",
                              address, mem[0],mem[1],mem[2],mem[3],mem[4],mem[5],mem[6],mem[7]);
        WriteLine(line);
        address += 8;
      } /* if */
      else {
        break;
      } /* else */
    } /* for */
  } /* if */

  /* display exact the requuested values */
  while ((result == BDI_OKAY && count--)) {
    result = TARGET_ReadByte(address, &mem[0]);
    if (result == BDI_OKAY) {
      (void)sprintf(line, "%08lx : 0x%02x  % 4i", address, mem[0], mem[0]);
      WriteLine(line);
      address++;
    } /* if */
  } /* while */

  nextTargetAddr = address;
  return result;
} /* DisplayByte */


/****************************************************************************
 ****************************************************************************

    DisplayGPR :

    Display general purpose registers.

    INPUT  : -
    OUTPUT : RETURN     error code

 ****************************************************************************/

static int DisplayGPR(void)
{
  int   i;
  int   j;
  int   result;
  char  line[80];
  DWORD gpr[4];
  
  static BDI_RegTypeT regList[] = {BDI_RT_ARM_CPSR, 0};

  /* display gereral purpose registers */
  result = BDI_OKAY;
  for (i=0; i<4; i++) {
      for (j=0; j<4; j++) {
        if (result == BDI_OKAY) result = TARGET_ReadGPR((WORD)((4 * i) + j), &gpr[j]);
      } /* for */
    if (result == BDI_OKAY) {
      (void)sprintf(line, "GPR%02i: %08lx %08lx %08lx %08lx", (4*i), gpr[0], gpr[1], gpr[2], gpr[3]);
      WriteLine(line);
    } /* if */
    else {
      break;
    } /* else */
  } /* for */

  /* display PSR and PC */
  if (result == BDI_OKAY) result = BDI_RegisterGet(1, regList, gpr);
  if (result == BDI_OKAY) {
    (void)sprintf(line, "CPSR : %08lx", gpr[0]);
    WriteLine(line);
  } /* if */

  return result;
} /* DisplayGPR */


/****************************************************************************
 ****************************************************************************

    DisplayICEBreaker :

    Display OnCE registers.

    INPUT  : -
    OUTPUT : RETURN     error code

 ****************************************************************************/

static int DisplayICEBreaker(void)
{
  int   result;
  char  line[80];
  DWORD reg[17];

  static BDI_RegTypeT regList[] = 
  {
    { BDI_RT_ARM_ICEB,  0 },   // Debug Control
    { BDI_RT_ARM_ICEB,  1 },   // Debug Status
    { BDI_RT_ARM_ICEB,  4 },   // Debug Comms Control
    { BDI_RT_ARM_ICEB,  5 },   // Debug Comms Data
    { BDI_RT_ARM_ICEB,  8 },   // Watchpoint 0 Address Value
    { BDI_RT_ARM_ICEB,  9 },   // Watchpoint 0 Address Mask
    { BDI_RT_ARM_ICEB, 10 },   // Watchpoint 0 Data Value
    { BDI_RT_ARM_ICEB, 11 },   // Watchpoint 0 Data Mask
    { BDI_RT_ARM_ICEB, 12 },   // Watchpoint 0 Control Value
    { BDI_RT_ARM_ICEB, 13 },   // Watchpoint 0 Control Mask
    { BDI_RT_ARM_ICEB, 16 },   // Watchpoint 1 Address Value
    { BDI_RT_ARM_ICEB, 17 },   // Watchpoint 1 Address Mask
    { BDI_RT_ARM_ICEB, 18 },   // Watchpoint 1 Data Value
    { BDI_RT_ARM_ICEB, 19 },   // Watchpoint 1 Data Mask
    { BDI_RT_ARM_ICEB, 20 },   // Watchpoint 1 Control Value
    { BDI_RT_ARM_ICEB, 21 },   // Watchpoint 1 Control Mask
    { BDI_RT_ARM_ICEB,  2 }    // Vector Catch
  };
  
  result = BDI_RegisterGet(17 , regList, reg);
  if (result == BDI_OKAY) {
    (void)sprintf(line, "DBGC: %08lx     DBGS: %08lx", reg[0], reg[1]);
    WriteLine(line);
    (void)sprintf(line, "COMC: %08lx     COMD: %08lx", reg[2], reg[3]);
    WriteLine(line);

    (void)sprintf(line, "W0AV: %08lx     W0AM: %08lx", reg[4], reg[5]);
    WriteLine(line);
    (void)sprintf(line, "W0DV: %08lx     W0DM: %08lx", reg[6], reg[7]);
    WriteLine(line);
    (void)sprintf(line, "W0CV: %08lx     W0CM: %08lx", reg[8], reg[9]);
    WriteLine(line);

    (void)sprintf(line, "W1AV: %08lx     W1AM: %08lx", reg[10], reg[11]);
    WriteLine(line);
    (void)sprintf(line, "W1DV: %08lx     W1DM: %08lx", reg[12], reg[13]);
    WriteLine(line);
    (void)sprintf(line, "W1CV: %08lx     W1CM: %08lx", reg[14], reg[15]);
    WriteLine(line);

    (void)sprintf(line, "VCATCH: %08lx", reg[16]);
    WriteLine(line);

  } /* if */

  return result;
} /* DisplayICEBreaker */


/****************************************************************************
 ****************************************************************************

    DisplayCP15 :

    Display general purpose registers.

    INPUT  : 
    OUTPUT : RETURN     error code

 ****************************************************************************/

static int DisplayCP15(WORD regNbr)
{
  int           result;
  char          line[80];
  DWORD         value;
  BDI_RegTypeT  reg;

  reg.regType = BDI_RT_ARM_CP15;
  reg.regAddr = regNbr;
  result = BDI_RegisterGet(1, &reg, &value);
  if (result == BDI_OKAY) {
    (void)sprintf(line, "CP15 : %08lx", value);
    WriteLine(line);
  } /* if */

  return result;
} /* DisplayCP15 */


/****************************************************************************
 ****************************************************************************

  TARGET_LoadELF:

  Load an ELF file.

  INPUT  : fileName     the file name on the host
  OUTPUT : startAddr    the start address or the lowest loaded address
           errorAddr    the address of the failing memory
           RETURN       error code

 ****************************************************************************/

int TARGET_LoadELF(char* fileName, DWORD* startAddr, DWORD* errorAddr)
{
  FILE*         file;
  int           loadResult;
  size_t        rxCount;
  BYTE          rxBlock[1024];
  size_t        loadSize;
  DWORD         loadAddr;
  Elf32_Ehdr    elfHeader;
  Elf32_Phdr    progHeader;
  BYTE*         getPtr;
  DWORD         startTime;
  DWORD         loadTime;
  DWORD         loadCount;
  char          info[80];


  /* open ELF file */
  file = fopen(fileName,"rb");
  if (file == NULL) return ERR_FILE_ACCESS;
  
  /* read ELF header */
  rxCount = fread(rxBlock, 1, sizeof elfHeader, file);
  getPtr = BDI_ExtractBlock(sizeof elfHeader.e_ident, elfHeader.e_ident, rxBlock);
  if (elfHeader.e_ident[5] == ELFDATA2LSB) {
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_type, getPtr);
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_machine, getPtr);
    getPtr = UTIL_ExtractLongLE(&elfHeader.e_version, getPtr);
    getPtr = UTIL_ExtractLongLE(&elfHeader.e_entry, getPtr);
    getPtr = UTIL_ExtractLongLE(&elfHeader.e_phoff, getPtr);
    getPtr = UTIL_ExtractLongLE(&elfHeader.e_shoff, getPtr);
    getPtr = UTIL_ExtractLongLE(&elfHeader.e_flags, getPtr);
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_ehsize, getPtr);
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_phentsize, getPtr);
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_phnum, getPtr);
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_shentsize, getPtr);
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_shnum, getPtr);
    getPtr = UTIL_ExtractWordLE(&elfHeader.e_shstrndx, getPtr);
  } /* if */
  else {
    getPtr = BDI_ExtractWord(&elfHeader.e_type, getPtr);
    getPtr = BDI_ExtractWord(&elfHeader.e_machine, getPtr);
    getPtr = BDI_ExtractLong(&elfHeader.e_version, getPtr);
    getPtr = BDI_ExtractLong(&elfHeader.e_entry, getPtr);
    getPtr = BDI_ExtractLong(&elfHeader.e_phoff, getPtr);
    getPtr = BDI_ExtractLong(&elfHeader.e_shoff, getPtr);
    getPtr = BDI_ExtractLong(&elfHeader.e_flags, getPtr);
    getPtr = BDI_ExtractWord(&elfHeader.e_ehsize, getPtr);
    getPtr = BDI_ExtractWord(&elfHeader.e_phentsize, getPtr);
    getPtr = BDI_ExtractWord(&elfHeader.e_phnum, getPtr);
    getPtr = BDI_ExtractWord(&elfHeader.e_shentsize, getPtr);
    getPtr = BDI_ExtractWord(&elfHeader.e_shnum, getPtr);
    getPtr = BDI_ExtractWord(&elfHeader.e_shstrndx, getPtr);
  } /* else */
  if (    (rxCount              != sizeof elfHeader)
       || (elfHeader.e_ident[0] != ELFMAG0)
       || (elfHeader.e_ident[1] != ELFMAG1)
       || (elfHeader.e_ident[2] != ELFMAG2)
       || (elfHeader.e_ident[3] != ELFMAG3)
       || (elfHeader.e_type     != ET_EXEC)
     ) {
    return ERR_FILE_ACCESS;
  } /* if */

  /* download all physical segments */
  startTime  = GetTickCount();
  loadCount  = 0;
  loadResult = BDI_OKAY;
  while ((elfHeader.e_phnum > 0) && (loadResult == BDI_OKAY)) {

    /* read program header */
    fseek(file, elfHeader.e_phoff, SEEK_SET);
    rxCount = fread(rxBlock, 1, sizeof progHeader, file);
    if (elfHeader.e_ident[5] == ELFDATA2LSB) {
      getPtr = UTIL_ExtractLongLE(&progHeader.p_type, rxBlock);
      getPtr = UTIL_ExtractLongLE(&progHeader.p_offset, getPtr);
      getPtr = UTIL_ExtractLongLE(&progHeader.p_vaddr, getPtr);
      getPtr = UTIL_ExtractLongLE(&progHeader.p_paddr, getPtr);
      getPtr = UTIL_ExtractLongLE(&progHeader.p_filesz, getPtr);
      getPtr = UTIL_ExtractLongLE(&progHeader.p_memsz, getPtr);
      getPtr = UTIL_ExtractLongLE(&progHeader.p_flags, getPtr);
      getPtr = UTIL_ExtractLongLE(&progHeader.p_align, getPtr);
    } /* if */
    else {
      getPtr = BDI_ExtractLong(&progHeader.p_type, rxBlock);
      getPtr = BDI_ExtractLong(&progHeader.p_offset, getPtr);
      getPtr = BDI_ExtractLong(&progHeader.p_vaddr, getPtr);
      getPtr = BDI_ExtractLong(&progHeader.p_paddr, getPtr);
      getPtr = BDI_ExtractLong(&progHeader.p_filesz, getPtr);
      getPtr = BDI_ExtractLong(&progHeader.p_memsz, getPtr);
      getPtr = BDI_ExtractLong(&progHeader.p_flags, getPtr);
      getPtr = BDI_ExtractLong(&progHeader.p_align, getPtr);
    } /* else */
    if (rxCount != sizeof progHeader) {
      fclose(file);
      return ERR_FILE_ACCESS;
    } /* if */
    elfHeader.e_phnum--;
    elfHeader.e_phoff += elfHeader.e_phentsize;

    /* abort if not a ROM section */
    if (   (progHeader.p_type != PT_LOAD) 
        || (    ((progHeader.p_flags & PF_X) == 0) 
             && ((progHeader.p_flags & PF_R) == 0)
           )
        ) {
      continue;
    } // if

    /* move to begin of program start */
    fseek(file, progHeader.p_offset, SEEK_SET);

    /* download program segment */
    loadResult = BDI_OKAY;
    loadSize   = progHeader.p_filesz;
    loadAddr   = progHeader.p_vaddr;
    while ((loadSize > 0) && (loadResult == BDI_OKAY)) {
      rxCount = fread(rxBlock, 1, sizeof rxBlock, file);
      if (rxCount > 0) {
        if ((DWORD)rxCount > loadSize) rxCount = loadSize;
        loadResult = BDI_LoadBlock(loadAddr, (WORD)rxCount, rxBlock, errorAddr);
        loadSize  -= rxCount;
        loadAddr  += (DWORD)rxCount;
      } /* if */
      else {
        loadResult = ERR_FILE_ACCESS;
      } /* else if */
    } /* while */
    loadCount += progHeader.p_filesz;
  } /* while */

  loadTime = GetTickCount() - startTime;
  if (loadTime == 0) loadTime = 1;
  if (loadResult == BDI_OKAY)
  {
    sprintf(info, "Loading of %lu bytes passed (%lu bytes/s).",
                   loadCount, 
                   (1000L * loadCount) / loadTime);
    WriteLine(info);
  } /* if */
  
  fclose(file);
  
  *startAddr = elfHeader.e_entry;
  return loadResult;
} /* TARGET_LoadELF */


/****************************************************************************
 ****************************************************************************

    Erase/Program EB40A flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListEB40A[] = {
  {BDI_INIT_ARM_WGPR, 8003,       0         },  // Successor count
  {BDI_INIT_ARM_WGPR, 8004,       0         },  // Successor size
  {BDI_INIT_ARM_WM32, 0xFFE00020, 0x00000001},  // Cancel reset remapping
  {BDI_INIT_ARM_WM32, 0xFFE00024, 0x00000006},  // enable A0..A20
  {BDI_INIT_ARM_WM32, 0xFFE00000, 0x01002539},  // Flash at 0x01000000
  {BDI_INIT_ARM_WM32, 0xFFFFF124, 0xFFFFFFFF},  // disable all interrupts
};

static int ProgramEB40A(void)
{
  int   result;
  DWORD errorAddr;

  result = BDI_FlashSetType(BDI_FLASH_AT49X16, 
                            0x200000, 
                            0,
                            16,
                            0x00000000);

  result = BDI_TargetStartup(100,      // 100ms reset time
                             1,        // ARM7TDMI, 16MHz JTAG Clock
                             sizeof initListEB40A / sizeof initListEB40A[0],
                             initListEB40A);
  

  result = BDI_FlashEraseSector(0x01140000);
  result = BDI_FlashEraseSector(0x01150000);
  result = BDI_FlashEraseSector(0x01160000);
  result = BDI_FlashEraseSector(0x01170000);
  result = BDI_FlashEraseSector(0x01180000);
  result = BDI_FlashWriteFile("e:\\temp\\eb40a.sss", &errorAddr);
  return result;
} /* ProgramEB40A */


/****************************************************************************
 ****************************************************************************

    Erase/Program SMDK40100 flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListSMDK40100[] = {
  {BDI_INIT_ARM_WGPR,	8006,	    1000      },  // Wake-up delay after releasing reset
  {BDI_INIT_ARM_WCPSR,	0,          0x000000D3},  // CPSR
  {BDI_INIT_ARM_WM16,	0x07FFA002,	0xA500    },  // BTCON
  {BDI_INIT_ARM_WM32,	0x07FF1000,	0x00017FF0},  // SYSCFG
  {BDI_INIT_ARM_WM8,	0x07FFD003,	0x18      },  // SYSCON
  {BDI_INIT_ARM_WM16,	0x07FFB012,	0x0707    },  // PCON1: Enable A16-A18
  {BDI_INIT_ARM_WM16,	0x07FFB016,	0x2AAA    },  // PCON3: enable WE0
  {BDI_INIT_ARM_WM16,	0x07FFB018,	0xFFFF    },  // PCON4: enable D8-D15
  {BDI_INIT_ARM_WM32,	0x07FF2000,	0x02000071},  // BANKCON0
  {BDI_INIT_ARM_WM32,	0x07FF2004,	0x00000000},  // BANKCON1
  {BDI_INIT_ARM_WM32,	0x07FF2008,	0x00000000},  // BANKCON2
  {BDI_INIT_ARM_WM32,	0x07FF200C,	0x00000000},  // BANKCON3
  {BDI_INIT_ARM_WM32,	0x07FF2010,	0x00000000},  // BANKCON4
  {BDI_INIT_ARM_WM32,	0x07FF2014,	0x00000000},  // BANKCON5
  {BDI_INIT_ARM_WM32,	0x07FF2018,	0x00000000},  // BANKCON6
  {BDI_INIT_ARM_WM32,	0x07FF201C,	0x00000000},  // BANKCON7
  {BDI_INIT_ARM_WM32,	0x07FF2020,	0x00000001},  // REFCON
};

static BYTE flashData[] = {1,2,3,4,5,6,7,8};

static int ProgramSMDK40100(void)
{
  int   result;
  DWORD errorAddr;

  result = BDI_FlashSetType(BDI_FLASH_AM29FX00W, 
                            0x100000, 
                            0,
                            16,
                            0x07FF0000);

  result = BDI_TargetStartup(100,      // 100ms reset time
                             (1L << 16) | (0L << 8) | 1,  // big endian, ARM7TDMI, 16MHz JTAG
                             sizeof initListSMDK40100 / sizeof initListSMDK40100[0],
                             initListSMDK40100);
  

  result = BDI_FlashEraseSector(0x10000);
/*
  result = BDI_FlashWriteBlock(0x10000, sizeof flashData, flashData, &errorAddr);
*/
  result = BDI_FlashWriteBinary("e:/temp/testdc.bin", 0x10000, &errorAddr);
  result = BDI_VerifyBinary("e:/temp/testdc.bin", 0x10000, &errorAddr);
  return result;
} /* ProgramSMDK40100 */


/****************************************************************************
 ****************************************************************************

    Erase/Program LPC2106 flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListLPC2106[] = {
  {BDI_INIT_ARM_WGPR,	8006,	    100       },  // Wake-up delay after releasing reset
  {BDI_INIT_ARM_WM32,	0xE01FC040,	0x00000001},  // MEMMAP: User flash mode
};


static int ProgramLPC2106(void)
{
  int   result;
  DWORD errorAddr;


  /* reset and init target */
  result = BDI_TargetStartup(500,      // 500ms reset time
                             (0L << 16) | (0L << 8) | 4,  // little endian, ARM7TDMI, 1 MHz JTAG
                             sizeof initListLPC2106 / sizeof initListLPC2106[0],
                             initListLPC2106);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetupLPC2000(0x20000,      // 128k flash device
                                 14745,        // fsys = 14.745MHz 
                                 0x40001000);  // internal SRAM for buffer, code and stack
  if (result != BDI_OKAY) return result;

  /* erase flash  */
  printf("erasing ...   ");
  result = BDI_FlashEraseSector(0x000000F0);  // erase sectors 4 - 7
  // result = BDI_FlashEraseSector(0x00007FFF);  // erase sectors 0 - 14
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  /* program/verify from a file */
  printf("programming...");
  result = BDI_FlashWriteBinary("e:/temp/dump16k.bin", 0x08000, &errorAddr);
  //result = BDI_FlashWriteBinary("E:/cygwin/home/bdidemo/arm/lpc2100.bin", 0x00000, &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("verifing...   ");
  result = BDI_VerifyBinary("e:/temp/dump16k.bin", 0x08000, &errorAddr);
  // result = BDI_VerifyBinary("E:/cygwin/home/bdidemo/arm/lpc2100.bin", 0x00000, &errorAddr);
  printf("passed\n");

  return result;
} /* ProgramLPC2106 */


/****************************************************************************
 ****************************************************************************

    Erase/Program LPC2129 flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListLPC2129[] = {
  {BDI_INIT_ARM_WGPR,	8006,	    100       },  // Wake-up delay after releasing reset
  {BDI_INIT_ARM_WM32,	0xE01FC040,	0x00000001},  // MEMMAP: User flash mode
  // Disable PLL for flash programming
  {BDI_INIT_ARM_WM32,	0xE01FC080,	0x00000000},  // PLLCON
  {BDI_INIT_ARM_WM32,	0xE01FC084,	0x00000000},  // PLLCFG
  {BDI_INIT_ARM_WM32,	0xE01FC08C,	0x000000AA},  // PLLFEED
  {BDI_INIT_ARM_WM32,	0xE01FC08C,	0x00000055},  // PLLFEED
};



static int ProgramLPC2129(void)
{
  int   result;
  DWORD errorAddr;


  /* reset and init target */
  result = BDI_TargetStartup(500,      // 500ms reset time
                             (0L << 16) | (0L << 8) | 4,  // little endian, ARM7TDMI, 1 MHz JTAG
                             sizeof initListLPC2129 / sizeof initListLPC2129[0],
                             initListLPC2129);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetupLPC2000(0x40000,      // 256k flash device
                                 12000,        // fsys = 12.000MHz 
                                 0x40001000);  // internal SRAM for buffer, code and stack
  if (result != BDI_OKAY) return result;

  /* erase flash  */
  printf("erasing ...   ");
  result = BDI_FlashEraseSector(0x0001FFFF);  // erase sectors 0 - 16
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  /* program/verify from a file */
  printf("programming...");
  result = BDI_FlashWriteFile("D:/abatron/bdi360/arm/pro/lpc248k.sss", &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("verifing...   ");
  result = BDI_VerifyFile("D:/abatron/bdi360/arm/pro/lpc248k.sss", &errorAddr);
  printf("passed\n");

  return result;
} /* ProgramLPC2129 */


/****************************************************************************
 ****************************************************************************

    Erase/Program ADuC7000 flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListADuC7020[] = {
  {BDI_INIT_ARM_WGPR,	8006,	    300       },  // Wake-up delay after releasing reset
  {BDI_INIT_ARM_WGPR,	8008,	    0x00010000},  // Set save debug PC to internal SRAM
  {BDI_INIT_ARM_WCPSR,	0,	        0x000000D3},  // CPSR: set superviser mode
};


static int ProgramADuC7020(void)
{
  int   result;
  DWORD errorAddr;
  // DWORD sector;

  /* reset and init target */
  result = BDI_TargetStartup(100,      // 100ms reset time
                             (0L << 16) | (0L << 8) | 1,  // little endian, ARM7TDMI, 16 MHz JTAG
                             sizeof initListADuC7020 / sizeof initListADuC7020[0],
                             initListADuC7020);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetType(BDI_FLASH_ADUC7000, 
                            0x10000, 
                            0,
                            16,
                            0x00010020);

  if (result != BDI_OKAY) return result;

  /* erase flash  */
  /*
  for (sector = 0x80000; sector <= 0x8F600; sector += 0x200) {
    printf("erasing sector 0x%08lx ...   ", sector);
    result = BDI_FlashEraseSector(sector);
    if (result != BDI_OKAY) return result;
    printf("passed\n");
  }
  */
  printf("erasing all sectors ...  ");
  result = BDI_FlashErase(BDI_ERASE_CHIP, 0x80000);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  /* program/verify from a file */
  printf("programming...");
  result = BDI_FlashWriteFile("D:/abatron/bdi360/arm/pro/aduc7020.sss", &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("verifing...   ");
  result = BDI_VerifyFile("D:/abatron/bdi360/arm/pro/aduc7020.sss", &errorAddr);
  printf("passed\n");

  return result;
} /* ProgramADuC7020 */


/****************************************************************************
 ****************************************************************************

    Erase/Program STA2051 flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListSTA2051[] = {
  {BDI_INIT_ARM_WGPR,	8006,	    300       },  // Wake-up delay after releasing reset
  {BDI_INIT_ARM_WGPR,	8008,	    0x20000000},  // Set save debug PC to internal SRAM
  {BDI_INIT_ARM_WGPR,	15,         0x00000000},  // Set PC to 0x00000000
  {BDI_INIT_ARM_WCPSR,	0,	        0x000000D3},  // CPSR: set superviser mode
};


static int ProgramSTA2051(void)
{
  int   result;
  DWORD errorAddr;

  /* reset and init target */
  result = BDI_TargetStartup(500,   // 500ms reset time
                             (0L << 16) | (0L << 8) | 2,  // little endian, ARM7TDMI, 8 MHz JTAG
                             sizeof initListSTA2051 / sizeof initListSTA2051[0],
                             initListSTA2051);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetType(BDI_FLASH_STA2051, 
                            0x40000, 
                            0,
                            32,
                            0x20000000);
  if (result != BDI_OKAY) return result;

  /* erase flash  */
  /*
  printf("erasing Bank 0 : Sector 0 - 4 ...  ");
  result = BDI_FlashEraseSector(0x0000001f);
  if (result != BDI_OKAY) return result;
  printf("passed\n");
  printf("erasing Bank 0 : Sector 5     ...  ");
  result = BDI_FlashEraseSector(0x00000020);
  if (result != BDI_OKAY) return result;
  printf("passed\n");
  printf("erasing Bank 0 : Sector 6     ...  ");
  result = BDI_FlashEraseSector(0x00000040);
  if (result != BDI_OKAY) return result;
  printf("passed\n");
  printf("erasing Bank 0 : Sector 7     ...  ");
  result = BDI_FlashEraseSector(0x00000080);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("erasing Bank 1 : Sector 0 - 1 ...  ");
  result = BDI_FlashEraseSector(0x00030000);
  if (result != BDI_OKAY) return result;
  printf("passed\n");
  */

  printf("erasing Bank 0 : Sector 0 - 7 ...  ");
  result = BDI_FlashEraseSector(0x000000ff);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("erasing Bank 1 : Sector 0 - 1 ...  ");
  result = BDI_FlashEraseSector(0x00030000);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  /* reset and init target again becuase old flash code cuase programming error */
  result = BDI_TargetStartup(500,   // 500ms reset time
                             (0L << 16) | (0L << 8) | 2,  // little endian, ARM7TDMI, 8 MHz JTAG
                             sizeof initListSTA2051 / sizeof initListSTA2051[0],
                             initListSTA2051);
  if (result != BDI_OKAY) return result;
  
  /* program/verify from a file */
  printf("programming...");
  result = BDI_FlashWriteFile("D:/abatron/bdi360/arm/pro/sta2051_256k.sss", &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("verifing...   ");
  result = BDI_VerifyFile("D:/abatron/bdi360/arm/pro/sta2051_256k.sss", &errorAddr);
  printf("passed\n");

  return result;
} /* ProgramSTA2051 */


/****************************************************************************
 ****************************************************************************

    Erase/Program ST30F774 flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListST30F774[] = {
  {BDI_INIT_ARM_WGPR,	8009,	    1         },  // TRST driver type is push-pull
  {BDI_INIT_ARM_WGPR,	8007,	    2         },  // Use special ST30 reset sequence
  {BDI_INIT_ARM_WGPR,	8006,	    1000      },  // Delay for 1000 us after TRST/RST released
  {BDI_INIT_ARM_WGPR,	8008,	    0xA0000000},  // Set save debug PC to internal SRAM
  {BDI_INIT_ARM_WGPR,	15,         0x00000000},  // Set PC to 0x00000000
  {BDI_INIT_ARM_WCPSR,	0,	        0x000000D3},  // CPSR: set superviser mode
};


static int ProgramST30F774(void)
{
  int   result;
  DWORD errorAddr;

  /* reset and init target */
  result = BDI_TargetStartup(100,   // 100ms reset time
                             (0L << 16) | (0L << 8) | 1,  // little endian, ARM7TDMI, 16 MHz JTAG
                             sizeof initListST30F774 / sizeof initListST30F774[0],
                             initListST30F774);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetType(BDI_FLASH_ST30F,  // ST30F774
                            0x80000,          // 512K
                            0,
                            32,
                            0xA0000000);      // workspace in internal SRAM
  if (result != BDI_OKAY) return result;

  /* erase flash  */
  printf("erasing Bank 0 : Sector 0 - 11 ...  ");
  result = BDI_FlashEraseSector(0x00000fff);
  if (result != BDI_OKAY) return result;
  printf("passed\n");
  
  /* program/verify from a file */
  printf("programming...");
  result = BDI_FlashWriteFile("D:/abatron/bdi360/arm/pro/st30f_512k.sss", &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("verifing...   ");
  result = BDI_VerifyFile("D:/abatron/bdi360/arm/pro/st30f_512k.sss", &errorAddr);
  printf("passed\n");

  return result;
} /* ProgramST30F774 */


/****************************************************************************
 ****************************************************************************

    Erase/Program AT91SAM7S flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListAT91SAM7S[] = {
  {BDI_INIT_ARM_WM32,	0xFFFFFD44,	0x00008000},  // Disable watchdog
  {BDI_INIT_ARM_WM32,	0xFFFFFD08,	0xA5000001},  // Enable user reset
  {BDI_INIT_ARM_WM32,	0xFFFFFC20,	0x00000601},  // CKGR_MOR : Enabling the Main Oscillator
  {BDI_INIT_ARM_DELAY,	0,	        20	      },
  {BDI_INIT_ARM_WM32,	0xFFFFFC2C,	0x10480A0E},  // CKGR_PLLR: 96.1MHz (DIV=14,MUL=72+1)
  {BDI_INIT_ARM_DELAY,	0,	        20	      },
  {BDI_INIT_ARM_WM32,	0xFFFFFC30,	0x00000007},  // PMC_MCKR : MCK = PLL / 2 = 48MHz
  {BDI_INIT_ARM_DELAY,	0,	        20	      },
  {BDI_INIT_ARM_WGPR,	8005, 	    0x00000001},  // Set new JTAG clock to 16 MHz
  {BDI_INIT_ARM_WM32,	0xFFFFFF60,	0x00300100},  // MC_FMR: Flash mode (FWS=1,FMCN=48)
};


static int CALLBACK DisplayProgress(DWORD count)
{
  printf("Done %li\r", count);
  return 0;
} /* DisplayProgress */



static int ProgramAT91SAM7S(void)
{
  int   result;
  DWORD errorAddr;

  /* install callback procedure */
  BDI_InstallCallback(DisplayProgress);

  /* reset and init target */
  result = BDI_TargetStartup(100,   // 100ms reset time
                             (0L << 16) | (0L << 8) | 10,  // little endian, ARM7TDMI, 10kHz JTAG
                             sizeof initListAT91SAM7S / sizeof initListAT91SAM7S[0],
                             initListAT91SAM7S);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetType(BDI_FLASH_AT91SAM7S,  // AT91SAM7S64K
                            0x10000,              // 64K
                            0,
                            32,
                            0xFFFFFFFF);
  if (result != BDI_OKAY) return result;

  /* erase all  */
  printf("erasing all ...  ");
  result = BDI_FlashErase(BDI_ERASE_CHIP, 0x00100000);
  if (result != BDI_OKAY) return result;
  printf("passed\n");
  
  /* program/verify from a file */
  printf("programming...\n");
  result = BDI_FlashWriteFile("D:/abatron/bdi360/arm/pro/sam7s_64k.sss", &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("\npassed\n");

  printf("verifing...   \n");
  result = BDI_VerifyFile("D:/abatron/bdi360/arm/pro/sam7s_64k.sss", &errorAddr);
  printf("\npassed\n");

  return result;
} /* ProgramAT91SAM7S */


/****************************************************************************
 ****************************************************************************

    Erase/Program STR912F flash :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListSTR912[] = {
  {BDI_INIT_ARM_WGPR,	8006,	    250       },  // Wakeup time after reset released
  {BDI_INIT_ARM_WGPR,	8008,	    0x40000000},  // Set save debug PC to internal SRAM
  {BDI_INIT_ARM_WGPR,	8001,	    1         },  // Number of JTAG devices before the ARM core
  {BDI_INIT_ARM_WGPR,	8002,	    5         },  // Total IR length before the ARM core
  {BDI_INIT_ARM_WGPR,	8003,	    1         },  // Number of JTAG devices after the ARM core
  {BDI_INIT_ARM_WGPR,	8004,	    8         },  // Total IR length after the ARM core
  // Set clock to 48MHz via PLL
  {BDI_INIT_ARM_WM32,	0x5C002004,	0x000BC019},  // SCU_PLLCONF: PLL=ENA,P=3,N=C0,M=19
  {BDI_INIT_ARM_DELAY,	0,	        100       },  // let PLL lock
  {BDI_INIT_ARM_WM32,	0x5C002000,	0x00020000},  // SCU_CLKCNTR: MCLK=PLL
  // Setup flash memory interface (FMI)
  {BDI_INIT_ARM_WM32,	0x54000000,	0x00000004},  // FMI_BBSR  : boot bank size = 512k
  {BDI_INIT_ARM_WM32,	0x5400000C,	0x00000000},  // FMI_BBADR : boot bank addr = 0x00000000
  {BDI_INIT_ARM_WM32,	0x54000004,	0x00000002},  // FMI_NBBSR : boot bank size = 32k
  {BDI_INIT_ARM_WM32,	0x54000010,	0x00020000},  // FMI_NBBADR: boot bank addr = 0x00080000
  {BDI_INIT_ARM_WM32,	0x54000018,	0x00000018},  // FMI_CR    : enable both flash banks
  // load an execute code to set Configuration Control Register bit 18
  {BDI_INIT_ARM_WM32,	0x40000000,	0xEE3F0F11},  // mrc 15, 1, r0, cr15, cr1, {0}
  {BDI_INIT_ARM_WM32,	0x40000004,	0xE3800701},  // orr r0, r0, #262144
  {BDI_INIT_ARM_WM32,	0x40000008,	0xEE2F0F11},  // mcr 15, 1, r0, cr15, cr1, {0}
  {BDI_INIT_ARM_WM32,	0x4000000C,	0xEAFFFFFE},  // b   $pc
  {BDI_INIT_ARM_EXEC,	0x40000000,	0x00000000},  // execute code
};


static int ProgramSTR912(void)
{
  int   result;
  int   sector;
  DWORD addr;
  DWORD errorAddr;

  /* reset and init target */
  result = BDI_TargetStartup(500,   // 500ms reset time
                             (0L << 16) | (11L << 8) | 4,  // little endian, ARM966, 1 MHz JTAG
                             sizeof initListSTR912 / sizeof initListSTR912[0],
                             initListSTR912);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetType(BDI_FLASH_STR910F,
                            0x80000,          // 512K
                            0,
                            16,
                            0x40000000);      // workspace in internal SRAM
  if (result != BDI_OKAY) return result;

  /* unlock and erase sectors */
  addr = 0;
  for (sector = 0; sector < 8; sector++) {
    printf("erasing sector 0x%08lx ...   ", addr);
    result = BDI_SetWord(addr, 0x0060);
    result = BDI_SetWord(addr, 0x00D0);
    result = BDI_FlashEraseSector(addr);
    if (result != BDI_OKAY) return result;
    printf("passed\n");
    addr += 0x10000;
  } /* for */

  /* program/verify from a file */
  printf("programming...");
  result = BDI_FlashWriteBinary("e:/temp/dump512k.bin", 0x00000000, &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("verifing...   ");
  result = BDI_VerifyBinary("e:/temp/dump512k.bin", 0x00000000, &errorAddr);
  printf("passed\n");

  return result;
} /* ProgramSTR912 */


/****************************************************************************
 ****************************************************************************

    Erase/Program DaVinci :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListDaVinci[] = {
  {BDI_INIT_ARM_WGPR,	8009,	    1         },  // TRST driver type is push-pull
  {BDI_INIT_ARM_WGPR,	8006,	    100       },  // Wakeup time after reset released
  {BDI_INIT_ARM_WGPR,	8001,	    1         },  // Number of JTAG devices before the ARM core
  {BDI_INIT_ARM_WGPR,	8002,	    6         },  // Total IR length before the ARM core
  {BDI_INIT_ARM_WGPR,	8003,	    0         },  // Number of JTAG devices after the ARM core
  {BDI_INIT_ARM_WGPR,	8004,	    0         },  // Total IR length after the ARM core
  {BDI_INIT_ARM_WGPR,	8015,	    0x1F      },  // set vector catch mask during reset
};


// Configure ICEPick module to make ARM926 TAP visible
static const char* scanInit = 
"r0:w10000:"
"r1:t1:w1000:t0:w1000:" // assert reset and toggle TRST
"i6=07:d8=89:i6=02:"    // connect and select router
"d32=81000082:"         // set IP control
"d32=a018206f:"         // configure TAP0
"d32=a018216f:cl5:"     // enable TAP0, clock 5 times in RTI
"i10=ffff";             // scan bypass


static int ProgramDaVinci(void)
{
  int   result;

  /* download Scan Init string */
  result = BDI_SetOption(BDI_OPT_ARM_SCANINIT, scanInit);


  /* reset and init target */
  result = BDI_TargetStartup(500,      // 500ms reset time
                             (0L << 16) | (14L << 8) | 4,  // little endian, ARM926E, 1 MHz JTAG
                             sizeof initListDaVinci / sizeof initListDaVinci[0],
                             initListDaVinci);
  if (result != BDI_OKAY) return result;
  
  return result;
} /* ProgramDaVinci */



/****************************************************************************
 ****************************************************************************

    Startup/Program iMX31 :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListIMX31[] = {
  {BDI_INIT_ARM11_WGPR,	8006,	    250       },  // Wakeup time after reset released
  {BDI_INIT_ARM11_WGPR,	8001,	    2         },  // Number of JTAG devices before the ARM core
  {BDI_INIT_ARM11_WGPR,	8002,	    9         },  // Total IR length before the ARM core
  {BDI_INIT_ARM11_WGPR,	8003,	    1         },  // Number of JTAG devices after the ARM core
  {BDI_INIT_ARM11_WGPR,	8004,	    4         },  // Total IR length after the ARM core
  {BDI_INIT_ARM11_CP15,	0x0001,	    0x00050078},  // CP15 control register
  {BDI_INIT_ARM11_CP15,	0x0707,	    0x00000000},  // CP15 invalidate I&D
  {BDI_INIT_ARM11_CP15,	0x0708,	    0x00000000},  // CP15 invalidate TLB
  {BDI_INIT_ARM11_CP15,	0x4A07,	    0x00000000},  // CP15 clean write buf
};


static int StartupIMX31(void)
{
  int   result;

  /* reset and init target */
  result = BDI_TargetStartup(500,                         // 500ms reset time
                             (0L << 16) | (0L << 8) | 1,  // little endian, ARM1136, 16 MHz JTAG
                             sizeof initListIMX31 / sizeof initListIMX31[0],
                             initListIMX31);
  if (result != BDI_OKAY) return result;
  
  return result;
} /* StartupIMX31 */


static int ProgramIMX31(void)
{
  int   result;
  int   sector;
  DWORD addr;
  DWORD errorAddr;

  /* reset and init target */
  result = BDI_TargetStartup(500,                         // 500ms reset time
                             (0L << 16) | (0L << 8) | 2,  // little endian, ARM1136, 16 MHz JTAG
                             sizeof initListIMX31 / sizeof initListIMX31[0],
                             initListIMX31);
  if (result != BDI_OKAY) return result;
  
  /* setup flash type */
  result = BDI_FlashSetType(BDI_FLASH_I28BX16,
                            0x200000,          // 2MB
                            0,
                            16,
                            0x1fffc000);      // workspace in internal SRAM
  if (result != BDI_OKAY) return result;

  /* unlock and erase sectors */
  addr = 0xa0140000;
  for (sector = 0; sector < 4; sector++) {
    printf("erasing sector 0x%08lx ...   ", addr);
    result = BDI_SetWord(addr, 0x0060);
    result = BDI_SetWord(addr, 0x00D0);
    result = BDI_FlashEraseSector(addr);
    if (result != BDI_OKAY) return result;
    printf("passed\n");
    addr += 0x10000;
  } /* for */

  /* program/verify from a file */
  printf("programming...");
  result = BDI_FlashWriteBinary("e:/temp/dump256k.bin", 0xa0140000, &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  printf("verifying...   ");
  result = BDI_VerifyBinary("e:/temp/dump256k.bin", 0xa0140000, &errorAddr);
  printf("passed\n");

  return result;
} /* ProgamIMX31 */


/****************************************************************************
 ****************************************************************************

    Load and execute some code :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

#define TARGET_READY    1
#define TARGET_RUNNING  2


#define CODE_ADDR       0x40000000
#define DATA_ADDR       0x40010000
#define FLASH_ADDR      0x20000000


static BDI_InitTypeT initList926[] = {
  {BDI_INIT_ARM_WGPR,	8006,	    250       },  // Wakeup time after reset released

    // add additional configurations
    // setup SDRAM 
    // ............
};


static int ExecuteCode926(void)
{
  int   result;
  int   timeout;
  int   sector;
  DWORD addr;
  DWORD errorAddr;
  WORD      state;
  DWORD     timeStamp;
  WORD      infoCount;
  DWORD     infoValue[10];
  int       errorCode;

  static const BDI_RegTypeT regType[] = {
    { BDI_RT_ARM_GPR,  3 },  // maybe used to setup parameters
    { BDI_RT_ARM_GPR,  4 },
    { BDI_RT_ARM_GPR,  5 },
    { BDI_RT_ARM_GPR, 15 }   // sets the PC
  };

  static const DWORD regExec[] = {
    0x00100000L,  /* count      */
    DATA_ADDR,    /* source     */
    FLASH_ADDR,   /* flash addr */
    CODE_ADDR,    /* PC         */
  };


  /* reset and init target */
  result = BDI_TargetStartup(500,      // 500ms reset time
                             (0L << 16) | (14L << 8) | 4,  // little endian, ARM926E, 1 MHz JTAG
                             sizeof initList926 / sizeof initList926[0],
                             initList926);
  if (result != BDI_OKAY) return result;
  
  
  /* setup flash type (only used for erase) */
  result = BDI_FlashSetType(BDI_FLASH_STRATAX16,
                            0x02000000,      
                            0,
                            16,
                            0xFFFFFFFF);
  if (result != BDI_OKAY) return result;

  /* unlock and erase sectors */
  addr = FLASH_ADDR;
  for (sector = 0; sector < 8; sector++) {
    printf("erasing sector 0x%08lx ...   ", addr);
    result = BDI_SetWord(addr, 0x0060);
    result = BDI_SetWord(addr, 0x00D0);
    result = BDI_FlashEraseSector(addr);
    if (result != BDI_OKAY) return result;
    printf("passed\n");
    addr += 0x40000;
  } /* for */

  /* load the algorithm code into SDRAM */
  printf("loading code...");
  result = BDI_LoadBinary("code.bin", CODE_ADDR, &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  /* load the flash data into SDRAM */
  printf("loading data...");
  result = BDI_LoadBinary("image.bin", DATA_ADDR, &errorAddr);
  if (result != BDI_OKAY) return result;
  printf("passed\n");

  /* start algorithm code */
  result = BDI_RegisterSet(sizeof regType / sizeof regType[0], regType, regExec);
  result = BDI_TargetStart();
  timeout = 1000;
  do {
    Sleep(100);
    result = BDI_StateGet(&state, &errorCode, &timeStamp, &infoCount, infoValue);
  } while ((state == TARGET_RUNNING) && (--timeout > 0));

  /* read back for exampe some registres to check if programming was sucessful */


  return result;
} /* ExecuteCode926 */


/****************************************************************************
 ****************************************************************************

    Startup/Program iMX51 :

    INPUT  :
    OUTPUT : RETURN     error code

 ****************************************************************************/

static BDI_InitTypeT initListIMX51[] = {
  {BDI_INIT_ARM11_WGPR,	8006,	    100       },  // Wakeup time after reset released
  {BDI_INIT_ARM11_WGPR,	8001,	    2         },  // Number of JTAG devices before the ARM core
  {BDI_INIT_ARM11_WGPR,	8002,	    9         },  // Total IR length before the ARM core
  {BDI_INIT_ARM11_WGPR,	8003,	    0         },  // Number of JTAG devices after the ARM core
  {BDI_INIT_ARM11_WGPR,	8004,	    0         },  // Total IR length after the ARM core
  {BDI_INIT_ARM11_WGPR,	8014,	    0xe0008000},  // APB debug base address
};


// There is no debug communication possible when RESET is asserted.
// In order be able to hard reset the system we use SCANINIT
static const char* scanInitIMX51 = 
"r1:w100:t1:w100:t0:"   // assert reset and toggle TRST
"w1000:ch10:w1000:"     // clock TCK with TMS high and wait
"r0:w2000000";          // release reset and let run for 2 seconds


static int StartupIMX51(void)
{
  int   result;

  /* download Scan Init string */
  result = BDI_SetOption(BDI_OPT_ARM_SCANINIT, scanInitIMX51);

  /* reset and init target */
  result = BDI_TargetStartup(500,                         // 500ms reset time
                             (0L << 16) | (1L << 8) | 4,  // little endian, Cortex-A8, 8 MHz JTAG
                             sizeof initListIMX51 / sizeof initListIMX51[0],
                             initListIMX51);
  if (result != BDI_OKAY) return result;
  
  return result;
} /* StartupIMX51 */


/****************************************************************************
 ****************************************************************************
                iMX51 EVK SPI AT45 Data Flash Programming
 ****************************************************************************
 ****************************************************************************/

/* eCSPI register adresses */
#define SPI_RXDATA    0x70010000
#define SPI_TXDATA    0x70010004
#define SPI_CONTROL   0x70010008
#define SPI_CONFIG    0x7001000C
#define SPI_INT       0x70010010
#define SPI_DMA       0x70010014
#define SPI_STATUS    0x70010018
#define SPI_PERIOD    0x7001001C
#define SPI_TEST      0x70010020

#define SPI_CTRL_SS1  0x00047021  // control value for SS1 */
#define SPI_CTRL_XCH  (1L<<2)

#define SPI_STAT_TC   (1L<<7)
#define SPI_STAT_RO   (1L<<6)
#define SPI_STAT_RF   (1L<<5)
#define SPI_STAT_RDR  (1L<<4)
#define SPI_STAT_RR   (1L<<3)
#define SPI_STAT_TF   (1L<<2)
#define SPI_STAT_TDR  (1L<<1)
#define SPI_STAT_TE   (1L<<0)

#define AT45_PAGE_SIZE    528
#define AT45_STATUS_RDY   (1<<7)

static BYTE AT45_PageData[AT45_PAGE_SIZE];


/****************************************************************************
 ****************************************************************************

    AT45_ReadStatus :

 ****************************************************************************/

static BYTE AT45_ReadStatus(void)
{
  int       result;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     flashStatus;

  /* start a 2 byte SPI transfer sequence */
  spiControl = SPI_CTRL_SS1 | ((16-1) << 20);
  result = BDI_SetLong(SPI_CONTROL, spiControl);
  result = BDI_SetLong(SPI_TXDATA, 0xD700);
  result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
  result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);

  /* wait for SPI transfer completed */
  do {
    result = BDI_GetLong(SPI_STATUS, &spiStatus);
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* read data from RX Fifo and display */
  result = BDI_GetLong(SPI_RXDATA, &flashStatus);
  return (BYTE)flashStatus;
} /* AT45_ReadStatus */


/****************************************************************************
 ****************************************************************************

    AT45_DisplayFlashID :

 ****************************************************************************/

static int AT45_DisplayFlashID(void)
{
  int       result;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     flashID;

  /* start a 4 byte SPI transfer sequence */
  spiControl = SPI_CTRL_SS1 | ((32-1) << 20);
  result = BDI_SetLong(SPI_CONTROL, spiControl);
  result = BDI_SetLong(SPI_TXDATA, 0x9F000000);
  result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
  result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);

  /* wait for SPI transfer completed */
  do {
    result = BDI_GetLong(SPI_STATUS, &spiStatus);
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* read data from RX Fifo and display */
  result = BDI_GetLong(SPI_RXDATA, &flashID);
  printf("Flash Manu/ID: %06lx\n", flashID & 0x00ffffff);

  /* start a 2 byte SPI transfer sequence */
  spiControl = SPI_CTRL_SS1 | ((16-1) << 20);
  result = BDI_SetLong(SPI_CONTROL, spiControl);
  result = BDI_SetLong(SPI_TXDATA, 0xD700);
  result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
  result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);

  /* wait for SPI transfer completed */
  do {
    result = BDI_GetLong(SPI_STATUS, &spiStatus);
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* read data from RX Fifo and display */
  result = BDI_GetLong(SPI_RXDATA, &flashID);
  printf("Flash Status : %02lx\n", flashID);

  return result;
} /* AT45_DisplayFlashID */


/****************************************************************************
 ****************************************************************************

    AT45_PageRead :
    A page read is done in 4 steps of 132 bytes

 ****************************************************************************/

static int AT45_PageRead(DWORD page)
{
  int       result;
  int       part;
  int       i;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     spiCmdAddr;
  DWORD     flashData;
  DWORD     pageData[AT45_PAGE_SIZE / 4];
  DWORD*    pData;


  /*  Four 132+8 byte SPI transfer sequences */
  pData = pageData;
  spiCmdAddr = 0xD2000000 | (page << 10);
  spiControl = SPI_CTRL_SS1 | (((140*8)-1) << 20);
  for (part = 0; part < 4; part++) {
    result = BDI_SetLong(SPI_CONTROL, spiControl);
    result = BDI_SetLong(SPI_TXDATA, spiCmdAddr);   // page read command and address
    result = BDI_SetLong(SPI_TXDATA, 0);            // 4 dummy bytes
    for (i = 0; i < 33; i++) result = BDI_SetLong(SPI_TXDATA, 0);
    result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
    result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);
    spiCmdAddr += 132;

    /* wait for SPI transfer completed */
    do {
      result = BDI_GetLong(SPI_STATUS, &spiStatus);
    } while ((spiStatus & SPI_STAT_TC) == 0);

    /* read data from RX Fifo and store */
    (void)BDI_GetLong(SPI_RXDATA, &flashData);  // discard
    (void)BDI_GetLong(SPI_RXDATA, &flashData);  // discard
    for (i = 0; i < 33; i++) {
      (void)BDI_GetLong(SPI_RXDATA, pData);
      pData++;
    } /* for */
  } /* for */

  /* display page */
  pData = pageData;
  for (i=0; i < 33; i++) {
    (void)printf("%08lx%08lx%08lx%08lx\n",
                 *(pData+0), *(pData+1), *(pData+2), *(pData+3));
    pData += 4;
  } /* for */


  return result;
} /* AT45_PageRead */


/****************************************************************************
 ****************************************************************************

    AT45_PageProg :
    Program one page via Buffer Write and then Buffer to Memory Page

 ****************************************************************************/

static int AT45_PageProg(BOOL erase, DWORD page, const BYTE* data)
{
  int       result;
  int       part;
  int       i;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     spiCmdAddr;
  DWORD     flashData;
  BYTE      flashStatus;

  /*  write data to buffer 1 */
  spiCmdAddr = 0x84000000;
  spiControl = SPI_CTRL_SS1 | (((136*8)-1) << 20);
  for (part = 0; part < 4; part++) {
    result = BDI_SetLong(SPI_CONTROL, spiControl);
    result = BDI_SetLong(SPI_TXDATA, spiCmdAddr);
    for (i = 0; i < 33; i++) {
      flashData  = (DWORD)(*data++) << 24;
      flashData += (DWORD)(*data++) << 16;
      flashData += (DWORD)(*data++) << 8;
      flashData += (DWORD)(*data++);
      (void)BDI_SetLong(SPI_TXDATA, flashData);
    } /* for */
    result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
    result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);
    spiCmdAddr += 132;

    /* wait for SPI transfer completed */
    do {
      result = BDI_GetLong(SPI_STATUS, &spiStatus);
    } while ((spiStatus & SPI_STAT_TC) == 0);

    if (result != BDI_OKAY) return result;
  } /* for */

  /* program from buffer to main memory */
  spiCmdAddr  = (erase) ? 0x83000000 : 0x88000000;
  spiCmdAddr |= (page << 10);
  spiControl  = SPI_CTRL_SS1 | (((4*8)-1) << 20);
  result = BDI_SetLong(SPI_CONTROL, spiControl);
  result = BDI_SetLong(SPI_TXDATA, spiCmdAddr);
  result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
  result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);

  /* wait for SPI transfer completed */
  do {
    result = BDI_GetLong(SPI_STATUS, &spiStatus);
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* reset SPI in order to drop received data */
  result = BDI_SetLong(SPI_CONTROL, SPI_CTRL_SS1 & ~1L);

  /* wait until flash ready */
  do {
    flashStatus = AT45_ReadStatus();
  } while ((flashStatus & AT45_STATUS_RDY) == 0);

  return result;
} /* AT45_PageProg */


/****************************************************************************
 ****************************************************************************

    AT45_PageProgFast :
    Program one page via Buffer Write and then Buffer to Memory Page
    The Buffer Write is supported with code running on the target.

    Parameters:
    + 0 : Pointer to page data
    + 4 : Command for AT45
    + 8 : SPI control value
    +12 : Pointer to SPI TX Fifo
    +16 : Pointer to SPI Control Register
    +20 : Pointer to SPI Status Register

Helper code loaded to target RAM:
---------------------------------

typedef struct {
           BYTE*   data;
           UINT32  at45Cmd;
           UINT32  spiCtrl;
  volatile UINT32* pTxFifo;
  volatile UINT32* pControl;
  volatile UINT32* pStatus;
} ParamAT45T;

void AT45_BufferWrite(ParamAT45T* param)
{
  int       part;
  int       i;
  BYTE*     pData;
  DWORD     flashData;

  // write data to buffer 1
  pData = param->data;
  for (part = 0; part < 4; part++) {
    *param->pControl = param->spiCtrl;
    *param->pTxFifo  = param->at45Cmd;
    for (i = 0; i < 33; i++) {
      flashData  = (DWORD)(*pData++) << 24;
      flashData += (DWORD)(*pData++) << 16;
      flashData += (DWORD)(*pData++) << 8;
      flashData += (DWORD)(*pData++);
      *param->pTxFifo = flashData;
    }
    *param->pStatus  = (SPI_STAT_TC | SPI_STAT_RO);
    *param->pControl = param->spiCtrl | SPI_CTRL_XCH;
    param->at45Cmd  += 132;

    // wait for SPI transfer completed
    while ((*param->pStatus & SPI_STAT_TC) == 0);
  }
  __asm("bkpt");
}

 ****************************************************************************/

#define TARGET_WORKSPACE  0xa0001000
#define TARGET_PARAM      (TARGET_WORKSPACE + 0x000)
#define TARGET_CODE       (TARGET_WORKSPACE + 0x100)
#define TARGET_BUFFER     (TARGET_WORKSPACE + 0x400)

static const DWORD AT45_BufferWrite[] = {
0xe5903000, 	// a0008004:	ldr	r3, [r0]
0xe590a010, 	// a0008008:	ldr	sl, [r0, #16]
0xe590500c, 	// a000800c:	ldr	r5, [r0, #12]
0xe5907014, 	// a0008010:	ldr	r7, [r0, #20]
0xe1a06000, 	// a0008014:	mov	r6, r0
0xe3a08000, 	// a0008018:	mov	r8, #0	; 0x0
0xe3a090c0, 	// a000801c:	mov	r9, #192	; 0xc0
0xe5962008, 	// a0008020:	ldr	r2, [r6, #8]
0xe1a0c003, 	// a0008024:	mov	ip, r3
0xe2834084, 	// a0008028:	add	r4, r3, #132	; 0x84
0xe58a2000, 	// a000802c:	str	r2, [sl]
0xe5963004, 	// a0008030:	ldr	r3, [r6, #4]
0xe5853000, 	// a0008034:	str	r3, [r5]
0xe5dc3000, 	// a0008038:	ldrb	r3, [ip]
0xe5dc2001, 	// a000803c:	ldrb	r2, [ip, #1]
0xe5dc1002, 	// a0008040:	ldrb	r1, [ip, #2]
0xe5dc0003, 	// a0008044:	ldrb	r0, [ip, #3]
0xe1a03c03, 	// a0008048:	lsl	r3, r3, #24
0xe1a02802, 	// a000804c:	lsl	r2, r2, #16
0xe28cc004, 	// a0008050:	add	ip, ip, #4	; 0x4
0xe0833002, 	// a0008054:	add	r3, r3, r2
0xe154000c, 	// a0008058:	cmp	r4, ip
0xe0833000, 	// a000805c:	add	r3, r3, r0
0xe1a01401, 	// a0008060:	lsl	r1, r1, #8
0xe0833001, 	// a0008064:	add	r3, r3, r1
0xe5853000, 	// a0008068:	str	r3, [r5]
0x1afffff1, 	// a000806c:	bne	a0008038 <AT45_BufferWrite+0x38>
0xe5879000, 	// a0008070:	str	r9, [r7]
0xe5963008, 	// a0008074:	ldr	r3, [r6, #8]
0xe3833004, 	// a0008078:	orr	r3, r3, #4	; 0x4
0xe58a3000, 	// a000807c:	str	r3, [sl]
0xe5962004, 	// a0008080:	ldr	r2, [r6, #4]
0xe2822084, 	// a0008084:	add	r2, r2, #132	; 0x84
0xe5862004, 	// a0008088:	str	r2, [r6, #4]
0xe5973000, 	// a000808c:	ldr	r3, [r7]
0xe3130080, 	// a0008090:	tst	r3, #128	; 0x80
0x0afffffc, 	// a0008094:	beq	a000808c <AT45_BufferWrite+0x8c>
0xe2888001, 	// a0008098:	add	r8, r8, #1	; 0x1
0xe3580004, 	// a000809c:	cmp	r8, #4	; 0x4
0x11a03004, 	// a00080a0:	movne	r3, r4
0x1affffdd, 	// a00080a4:	bne	a0008020 <AT45_BufferWrite+0x20>
0xe1200070, 	// a00080a8:	bkpt	0x0000
};


static int LoadCode(DWORD addr, WORD count, const DWORD* code)
{
  int   result;
  WORD  i;
  DWORD readBack;

  /* copy code to target ram */
  for (i = 0; i < count; i++) {
    result = BDI_SetLong(addr, *code);
    if (result != BDI_OKAY) return result;
    result = BDI_GetLong(addr, &readBack);
    if (result != BDI_OKAY) return result;
    if (readBack != *code) return ERR_MEM_ACCESS;
    addr += 4;
    code++;
  } /* for */

  return BDI_OKAY;
} /* LoadCode */


static int AT45_LoadHelper(void)
{
  return LoadCode(TARGET_CODE, sizeof AT45_BufferWrite / 4, AT45_BufferWrite);
} /* AT45_LoadHelper */


static int AT45_PageProgFast(BOOL erase, DWORD page, BYTE* data)
{
  int       result;
  int       timeout;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     spiCmdAddr;
  DWORD     errorAddr;
  BYTE      flashStatus;
  WORD      state;
  DWORD     timeStamp;
  WORD      infoCount;
  DWORD     infoValue[10];
  int       errorCode;

  static const BDI_RegTypeT regTypeAT45[] = {
    { BDI_RT_ARM_GPR,  0 },  // function parameter
    { BDI_RT_ARM_GPR, 15 }   // sets the PC
  };

  static const DWORD regExecAT45[] = {
    TARGET_PARAM, // parameter block addr
    TARGET_CODE,  // PC
  };


  /* setup parameters */
  result = BDI_SetLong(TARGET_PARAM +  0, TARGET_BUFFER);
  result = BDI_SetLong(TARGET_PARAM +  4, 0x84000000);
  result = BDI_SetLong(TARGET_PARAM +  8, SPI_CTRL_SS1 | (((136*8)-1) << 20));
  result = BDI_SetLong(TARGET_PARAM + 12, SPI_TXDATA);
  result = BDI_SetLong(TARGET_PARAM + 16, SPI_CONTROL);
  result = BDI_SetLong(TARGET_PARAM + 20, SPI_STATUS);

  /* load page data to target workspace */
  result = BDI_LoadBlock(TARGET_BUFFER, AT45_PAGE_SIZE, data, &errorAddr);

  /* execute helper code */
  result = BDI_RegisterSet(sizeof regTypeAT45 / sizeof regTypeAT45[0], regTypeAT45, regExecAT45);
  result = BDI_TargetStart();
  timeout = 10000;
  do {
    result = BDI_StateGet(&state, &errorCode, &timeStamp, &infoCount, infoValue);
  } while ((state == TARGET_RUNNING) && (--timeout > 0));

  /* program from buffer to main memory */
  spiCmdAddr  = (erase) ? 0x83000000 : 0x88000000;
  spiCmdAddr |= (page << 10);
  spiControl  = SPI_CTRL_SS1 | (((4*8)-1) << 20);
  result = BDI_SetLong(SPI_CONTROL, spiControl);
  result = BDI_SetLong(SPI_TXDATA, spiCmdAddr);
  result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
  result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);

  /* wait for SPI transfer completed */
  do {
    result = BDI_GetLong(SPI_STATUS, &spiStatus);
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* reset SPI in order to drop received data */
  result = BDI_SetLong(SPI_CONTROL, SPI_CTRL_SS1 & ~1L);

  /* wait until flash ready */
  do {
    flashStatus = AT45_ReadStatus();
  } while ((flashStatus & AT45_STATUS_RDY) == 0);

  return result;
} /* AT45_PageProgFast */


/****************************************************************************
 ****************************************************************************

    AT45_PageErase  :    Erase one page
    AT45_BlockErase :    Erase one block
    AT45_ChipErase  :    Erase entire main memory

 ****************************************************************************/

static int AT45_PageErase(DWORD page)
{
  int       result;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     spiCmdAddr;
  DWORD     flashStatus;

  /* erase one page */
  spiCmdAddr = 0x81000000 | (page << 10);
  spiControl = SPI_CTRL_SS1 | (((4*8)-1) << 20);
  result = BDI_SetLong(SPI_CONTROL, spiControl);
  result = BDI_SetLong(SPI_TXDATA, spiCmdAddr);
  result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
  result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);

  /* wait for SPI transfer completed */
  do {
    result = BDI_GetLong(SPI_STATUS, &spiStatus);
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* reset SPI in order to drop received data */
  result = BDI_SetLong(SPI_CONTROL, SPI_CTRL_SS1 & ~1L);

  /* wait until page erased */
  do {
    flashStatus = AT45_ReadStatus();
  } while ((flashStatus & AT45_STATUS_RDY) == 0);

  return result;
} /* AT45_PageErase */


static int AT45_BlockErase(DWORD block)
{
  int       result;
  DWORD     spiControl;
  DWORD     spiStatus;
  DWORD     spiCmdAddr;
  DWORD     flashStatus;

  /* erase one block */
  spiCmdAddr = 0x50000000 | (block << 13);
  spiControl = SPI_CTRL_SS1 | (((4*8)-1) << 20);
  result = BDI_SetLong(SPI_CONTROL, spiControl);
  result = BDI_SetLong(SPI_TXDATA, spiCmdAddr);
  result = BDI_SetLong(SPI_STATUS, (SPI_STAT_TC | SPI_STAT_RO));
  result = BDI_SetLong(SPI_CONTROL, spiControl | SPI_CTRL_XCH);

  /* wait for SPI transfer completed */
  do {
    result = BDI_GetLong(SPI_STATUS, &spiStatus);
  } while ((spiStatus & SPI_STAT_TC) == 0);

  /* reset SPI in order to drop received data */
  result = BDI_SetLong(SPI_CONTROL, SPI_CTRL_SS1 & ~1L);

  /* wait until page erased */
  do {
    flashStatus = AT45_ReadStatus();
  } while ((flashStatus & AT45_STATUS_RDY) == 0);

  return result;
} /* AT45_BlockErase */


static int AT45_ChipErase(void)
{
  int       result;
  DWORD     block;

  /* because of the errata, don't use chip erase command */
  for (block = 0; block < 1024; block++) {
    result = AT45_BlockErase(block);
    if (result != BDI_OKAY) return result;
  } /* for */
  return BDI_OKAY;
} /* AT45_ChipErase */


/****************************************************************************
 ****************************************************************************

  AT45_ProgBinary:

  Program a binary file into the AT45 flash.
  512 byte of a page are from the file, the 16 spare bytes are:
  00,11,22,33,44,55,66,77,88,99,aa,bb,cc,dd,ee,ff

  INPUT  : startPage    the first page to program
           fileName     the binary file name
  OUTPUT : RETURN       error code

 ****************************************************************************/

static int AT45_ProgBinary(DWORD startPage, char* fileName)
{
  FILE*   file;
  int     result;
  int     count;
  BYTE    pageData[AT45_PAGE_SIZE];
  DWORD   progPage;
  DWORD   startTime;
  DWORD   progTime;
  DWORD   progCount;

  /* load buffer write helper code */
  result = AT45_LoadHelper();
  if (result != BDI_OKAY) return result;

  /* fill spare bytes */
  pageData[512] = 0x00;
  pageData[513] = 0x11;
  pageData[514] = 0x22;
  pageData[515] = 0x33;
  pageData[516] = 0x44;
  pageData[517] = 0x55;
  pageData[518] = 0x66;
  pageData[519] = 0x77;
  pageData[520] = 0x88;
  pageData[521] = 0x99;
  pageData[522] = 0xaa;
  pageData[523] = 0xbb;
  pageData[524] = 0xcc;
  pageData[525] = 0xdd;
  pageData[526] = 0xee;
  pageData[527] = 0xff;

  /* open binary file */
  file = fopen(fileName,"rb");
  if (file == NULL) return -1;

  /* program file */
  startTime = GetTickCount();
  progCount = 0;
  progPage  = startPage;
  do {
    memset(pageData, 0xff, 512);
    count = (int)fread(pageData, 1, 512, file);
    if (count > 0) {
      // result = AT45_PageProg(TRUE, progPage, pageData);
      result = AT45_PageProgFast(TRUE, progPage, pageData);
      progCount += count;
      progPage  += 1;
    } /* if */
  } while ((count > 0) && (result == BDI_OKAY));

  /* display programming performance */
  progTime = GetTickCount() - startTime;
  if (result == BDI_OKAY) {
    if (progTime == 0) progTime = 1;
    printf("Programming of %lu bytes passed (%lu bytes/s).\n",
           progCount,
           (1000L * progCount) / progTime);
  } /* if */

  fclose(file);

  return result;
} /* AT45_ProgBinary */


/****************************************************************************
 ****************************************************************************

  AT45_ProgRemote:

  Program via target code

  INPUT  : startPage    the first page to program
           fileName     the binary file name
  OUTPUT : RETURN       error code

 ****************************************************************************/

#define TGT_DATA_BUFFER   0xa0100000 

static int AT45_ProgRemote(DWORD startPage, char* fileName)
{
  int       result;
  int       timeout;
  FILE*     file;
  DWORD     fileSize;
  DWORD     startAddr;
  DWORD     errorAddr;
  WORD      state;
  int       errorCode;
  DWORD     timeStamp;
  WORD      infoCount;
  DWORD     infoValue[10];

  /* get binary file size */
  file = fopen(fileName,"rb");
  if (file == NULL) return -1;
  (void)fseek(file, 0, SEEK_END);
  fileSize = ftell(file);
  fclose(file);

  /* reset target */
  printf("Reset ... ");
  fflush(stdout);
  result = StartupIMX51();
  printf("passed\n");

  /* load data */
  printf("Loading data ... ");
  fflush(stdout);
  result = BDI_LoadBinary(fileName, TGT_DATA_BUFFER, &errorAddr);
  printf("passed\n");

  /* load programming algorithm */
  result = TARGET_LoadELF("m:/temp/helper.x", &startAddr, &errorAddr);

  /* set algorithm parameters */
  result = BDI_SetLong(0xa0007000, TGT_DATA_BUFFER);
  result = BDI_SetLong(0xa0007004, fileSize);
  result = BDI_SetLong(0xa0007008, startPage);

  /* execute target algorithm */
  printf("Programming ... ");
  fflush(stdout);
  result = TARGET_WriteGPR(15, startAddr);
  result = BDI_TargetStart();
  timeout = 10000;
  do {
    result = BDI_StateGet(&state, &errorCode, &timeStamp, &infoCount, infoValue);
  } while ((state == TARGET_RUNNING) && (--timeout > 0));
  printf("passed\n");

  return result;
} /* AT45_ProgRemote */


/****************************************************************************
 ****************************************************************************
                END of iMX51 EVK SPI AT45 Data Flash Programming
 ****************************************************************************
 ****************************************************************************/


/****************************************************************************
 ****************************************************************************

    ProcessCommand :

    Process a command line

    INPUT  : line       the command line
    OUTPUT : -

 ****************************************************************************/

BDI_InitTypeT initList[] = {
  {BDI_INIT_ARM_WM32, 0x0B000020, 0x00000000}  // Clear reset Map
};


static void ProcessCommand(char* line)
{
  static char* nameTable[] = {
     "MD","MDH","MDB","MM","MMH","MMB","RD","RDI","RDCP","RM","RMI","RMCP",
     "RESET","STARTUP","HALT","GO","TI","INFO","DUMP","PROG",
     "AT45ID","AT45READ","AT45PROG","AT45ERASE","AT45PROGFILE",
     "JTAG","HELP","EXIT"
  };

  typedef enum {
    N_MD,N_MDH,N_MDB,N_MM,N_MMH,N_MMB,N_RD,N_RDI,N_RDCP,N_RM,N_RMI,N_RMCP,
    N_RESET,N_STARTUP,N_HALT,N_GO,N_TI,N_INFO,N_DUMP,N_PROG,
    N_AT45ID,N_AT45READ,N_AT45PROG,N_AT45ERASE,N_AT45PROGFILE,
    N_JTAG,N_HELP,N_EXIT
  } NameT;

  char*                 linePtr;
  int                   command;
  int                   paramCount;
  DWORD                 param1;
  DWORD                 param2;
  DWORD                 param3;
  DWORD                 pc;
  char                  szParam[80];
  int                   result;
  char                  msg[256];
  DWORD                 errorParam;
  int                   i;

  /* extract command and parameters */
  paramCount = 0;
  param1     = 0;
  param2     = 0;
  param3     = 0;
  errorParam = 0;

  /* get command */
  linePtr = UTIL_MatchName(sizeof nameTable / sizeof nameTable[0],
                           nameTable,
                           &command,
                           line);

  /* get parameters */
  if (!ValidEndOfLine(linePtr)) {
    linePtr = UTIL_ExtractNumber(&param1, linePtr);
    paramCount++;
  } /* if */
  if (!ValidEndOfLine(linePtr)) {
    if ((command == N_AT45ERASE) || (command == N_AT45PROGFILE)) {
      linePtr = UTIL_ExtractString(szParam, linePtr);
    } /* if */
    else {
      linePtr = UTIL_ExtractNumber(&param2, linePtr);
    } /* else */
    paramCount++;
  } /* if */
  if (!ValidEndOfLine(linePtr)) {
    if (command == N_DUMP) {
      linePtr = UTIL_ExtractString(szParam, linePtr);
    } /* else if */
    else {
     linePtr = UTIL_ExtractNumber(&param3, linePtr);
    } /* else */
    paramCount++;
  } /* if */

  if ((command >= 0) && (ValidEndOfLine(linePtr))) {
    result = BDI_OKAY;
    switch (command) {

      case N_MD:
        if (paramCount == 0) param1 = nextTargetAddr;
        result = DisplayLong(param1, (WORD)param2);
        break;

      case N_MDH:
        if (paramCount == 0) param1 = nextTargetAddr;
        result = DisplayWord(param1, (WORD)param2);
        break;

      case N_MDB:
        if (paramCount == 0) param1 = nextTargetAddr;
        result = DisplayByte(param1, (WORD)param2);
        break;

      case N_MM:
        if (paramCount == 2)
          // result = TARGET_WriteLong(param1, param2);
          result = BDI_SetLong(param1, param2);
        else
          WriteLine("# syntax is MM <address> <value>");
        break;

      case N_MMH:
        if (paramCount == 2)
          result = TARGET_WriteWord(param1, (WORD)param2);
        else
          WriteLine("# syntax is MMH <address> <value>");
        break;

      case N_MMB:
        if (paramCount == 2)
          result = TARGET_WriteByte(param1, (BYTE)param2);
        else
          WriteLine("# syntax is MMB <address> <value>");
        break;

      case N_RD:
        result = DisplayGPR();
        break;

      case N_RDI:
        result = DisplayICEBreaker();
        break;

      case N_RDCP:
        result = DisplayCP15((WORD)param1);
        break;

      case N_RM:
        if (paramCount == 2)
          result = TARGET_WriteGPR((WORD)param1, param2);
        else
          WriteLine("# syntax is RM <number> <value>");
        break;

      case N_RMI:
        if (paramCount == 2)
          result = TARGET_WriteICEBreaker((WORD)param1, param2);
        else
          WriteLine("# syntax is RMI <number> <value>");
        break;

      case N_RMCP:
        if (paramCount == 2)
          result = TARGET_WriteCP15((WORD)param1, param2);
        else
          WriteLine("# syntax is RMCP <number> <value>");
        break;

      case N_RESET:
        result = BDI_TargetReset();
        break;

      case N_STARTUP:
        result = StartupIMX51();
        break;

      case N_HALT:
        result = BDI_TargetHalt();
        break;

      case N_GO:
        if (paramCount > 0) {
          result = TARGET_SetTargetPC(param1);
        } /* if */
        if (result == BDI_OKAY) {
          result = BDI_TargetStart();
        } /* if */
        break;

      case N_TI:
        if (paramCount > 0) {
          result = TARGET_SetTargetPC(param1);
        } /* if */
        result = BDI_TargetStep(0, 0, 0, &pc);
        if (result == BDI_OKAY) {
          (void)sprintf(msg, "Current PC: 0x%08lx", pc);
          WriteLine(msg);
        } /* if */
        break;

      case N_INFO:
        result = DisplayTargetInfo();
        break;

      case N_DUMP:
        result = BDI_DumpBinary(szParam, param1, param2);
        break;

      case N_PROG:
        result = ProgramIMX31();
        break;

      case N_AT45ID:
        result = AT45_DisplayFlashID();
        break;

      case N_AT45READ:
        result = AT45_PageRead(param1);
        break;

      case N_AT45PROG:
        for (i = 0; i < AT45_PAGE_SIZE; i++) AT45_PageData[i] = (BYTE)i;
        result = AT45_LoadHelper();
        result = AT45_PageProgFast(TRUE, param1, AT45_PageData);
        break;

      case N_AT45ERASE:
        if ((szParam[0] == 'c') || (szParam[0] == 'C'))
          result = AT45_ChipErase();
        if ((szParam[0] == 'b') || (szParam[0] == 'B'))
          result = AT45_BlockErase(param1);
        else
          result = AT45_PageErase(param1);
        break;

      case N_AT45PROGFILE:
        if (paramCount == 2)
          result = AT45_ProgBinary(param1, szParam);
          // result = AT45_ProgRemote(param1, szParam);
        else if (paramCount < 2)
          result = AT45_ProgBinary(param1, "e:/temp/dump256k.bin");
          // result = AT45_ProgRemote(param1, "e:/temp/dump256k.bin");
        else
          printf("# syntax is AT45PROGFILE <page> <filename>\n");
        break;
        break;

      case N_JTAG:
        result = BDI_DoJtag(JTAG_CMD_OPEN, 0, 0, NULL, NULL);
        if (result == BDI_OKAY) modeJTAG = TRUE;
        break;

      case N_HELP:
        DisplayHelp();
        break;

      case N_EXIT:
        terminate = TRUE;
        break;

      default:
        WriteLine("Sorry, command not jet implemented!");
        break;

    } /* switch */

    /* display error */
    if (result != BDI_OKAY) {
      DisplayError(result, errorParam);
    } /* if */
  } /* if */

  else {
    WriteLine("# syntax error in command");
  } /* else */

} /* ProcessCommand */




/****************************************************************************
 ****************************************************************************

    ProcessJTAG :

    Process a JTAG command line

    INPUT  : line       the command line
    OUTPUT : -

 ****************************************************************************/

#define JTAG_DATA_BYTES         32
#define JTAG_DATA_BITS          (8 * JTAG_DATA_BYTES)

/* Binary to ASCII-HEX conversion table */
static const char BinToHex[] = {'0','1','2','3','4','5','6','7','8','9',
                                'a','b','c','d','e','f'};

static char* jtagHelp[] = {
"TRST  {0|1}                  assert (1) or release (0) TRST",
"RST   {0|1}                  assert (1) or release (0) RST",
"SPEED <n>                    select JTAG clock frequency",
"CLK   <count> <tms>          clock TAP with requested TMS value",
"SCAN  <ctrl><len><...b2b1b0> scan with provided parameters",
"WIR   <len> <...b2b1b0>      write IR, b0 is first scanned",
"WDR   <len> <...b2b1b0>      write DR, b0 is first scanned",
"XIR   <len> <...b2b1b0>      xchg  IR, b0 is first scanned",
"XDR   <len> <...b2b1b0>      xchg  DR, b0 is first scanned",
"                             len : the number of bits 1..256",
"                             bx  : a data byte, two hex digits",
"HELP                         display JTAG command list",
"EXIT                         terminate JTAG mode"
};

static void JtagHelp(void)
{
  int   i;

  for (i=0; i< (sizeof jtagHelp / sizeof jtagHelp[0]); i++) {
    WriteLine(jtagHelp[i]);
  } /* for */
} /* JtagHelp */


static const char* ExtractHexByte(BYTE* value, const char* getPtr)
{
  BYTE  num;
  BYTE  digit;
  BYTE  c;
  int   i;

  while (*getPtr == ' ') getPtr++;
  num = 0;
  for (i = 0; i < 2; i++) {
    c = *getPtr;
    if      ((c >= '0') && (c <= '9')) digit = (BYTE)(c - '0');
    else if ((c >= 'a') && (c <= 'f')) digit = (BYTE)(c - 'a' + 10);
    else if ((c >= 'A') && (c <= 'F')) digit = (BYTE)(c - 'A' + 10);
    else break;
    num <<= 4;
    num += digit;
    getPtr++;
  } /* for */
  *value = num;
  return getPtr;
} /* ExtractHexByte */


static void ProcessJTAG(char* line)
{
  static char* jtagTable[] = {
     "TRST","RST","CLK","SPEED","SCAN","WIR","WDR","XIR","XDR","HELP","EXIT"
  };

  typedef enum {
    N_TRST,N_RST,N_CLK,N_SPEED,N_SCAN,N_WIR,N_WDR,N_XIR,N_XDR,N_HELP,N_EXIT
  } JtagT;

  const char*   linePtr;
  int           result;
  int           command;
  int           paramCount;
  DWORD         param1;
  DWORD         param2;
  DWORD         index;
  BYTE          dataWR[JTAG_DATA_BYTES];
  BYTE          dataRD[JTAG_DATA_BYTES];
  char          szData[2 * JTAG_DATA_BYTES + 1];
  char*         putPtr;

  /* extract command and parameters */
  paramCount = 0;
  param1     = 0;
  param2     = 0;
  linePtr = UTIL_MatchName(sizeof jtagTable / sizeof jtagTable[0],
                           jtagTable,
                           &command,
                           line);

  /* get 1. parameter */
  if (!ValidEndOfLine(linePtr)) {
    linePtr = UTIL_ExtractNumber(&param1, (char*)linePtr);
    paramCount++;
    if (   (command == N_WIR) || (command == N_WDR)
        || (command == N_XIR) || (command == N_XDR)) {
      if (param1 > JTAG_DATA_BITS) {
        WriteLine("# maximum data length is 256 bits");
        return;
      } /* if */
      if (param1 == 0) {
        WriteLine("# minimal data length is 1 bit");
        return;
      } /* if */
    } /* if */
  } /* if */

  /* get 2. parameter */
  if (!ValidEndOfLine(linePtr)) {
    if ((command == N_WIR) || (command == N_WDR) || (command == N_XIR) || (command == N_XDR)) {
      while (*linePtr == ' ') linePtr++;
      index = (param1 + 7) / 8;
      if (strlen(linePtr) != (2 * index)) {
        WriteLine("# data entry does not match length");
        return;
      } /* if */
      while (index > 0) {
        index--;
        linePtr = ExtractHexByte(&dataWR[index], linePtr);
      } /* while */
    } /* if */
    else {
      linePtr = UTIL_ExtractNumber(&param2, (char*)linePtr);
    } /* else */
    paramCount++;
    if (command == N_SCAN) {
      if (param2 > JTAG_DATA_BITS) {
        WriteLine("# maximum data length is 256 bits");
        return;
      } /* if */
      if (param2 == 0) {
        WriteLine("# minimal data length is 1 bit");
        return;
      } /* if */
    } /* if */
  } /* if */

  /* get 3. parameter */
  if (!ValidEndOfLine(linePtr)) {
    if (command == N_SCAN) {
      while (*linePtr == ' ') linePtr++;
      index = (param2 + 7) / 8;
      if (strlen(linePtr) != (2 * index)) {
        WriteLine("# data entry does not match length");
        return;
      } /* if */
      while (index > 0) {
        index--;
        linePtr = ExtractHexByte(&dataWR[index], linePtr);
      } /* while */
    } /* if */
    paramCount++;
  } /* if */

  /* process command */
  result = BDI_OKAY;
  if ((command >= 0) && (ValidEndOfLine(linePtr))) {
    switch (command) {

      case N_TRST:
        result = BDI_DoJtag(JTAG_CMD_PORT,
                            (BYTE)(((param1) ? JTAG_PORT_LOW : JTAG_PORT_HIGH) << 0),
                            0,
                            NULL,
                            NULL);
        break;

      case N_RST:
        result = BDI_DoJtag(JTAG_CMD_PORT,
                            (BYTE)(((param1) ? JTAG_PORT_LOW : JTAG_PORT_HIGH) << 2),
                            0,
                            NULL,
                            NULL);
        break;

      case N_CLK:
        if ((param1 >= 1) && (param1 <= 0xFFFF)) {
          result = BDI_DoJtag(JTAG_CMD_CLOCK, 
                              (BYTE)((param2) ? JTAG_CLOCK_TMS : 0), 
                              (WORD)param1, 
                              NULL, 
                              NULL);
        } /* if */
        else {
          WriteLine("# valid count is 1...64K");
        } /* else */
        break;

      case N_SPEED:
        if (param1 <= 16) {
          result = BDI_DoJtag(JTAG_CMD_SPEED, 
                              (BYTE)param1, 
                              0, 
                              NULL, 
                              NULL);
        } /* if */
        else {
          WriteLine("# valid speed is 0...16");
        } /* else */
        break;

      case N_SCAN:
        result = BDI_DoJtag(JTAG_CMD_SCAN,
                            (BYTE)param1,
                            (WORD)param2, 
                            dataWR, 
                            dataRD);
        break;

      case N_WIR:
      case N_WDR:
        result = BDI_DoJtag(JTAG_CMD_CLOCK, 0, 1, NULL, NULL);
        result = BDI_DoJtag((BYTE)((command == N_WIR) ? JTAG_CMD_IRSCAN : JTAG_CMD_DRSCAN),
                            0,
                            (WORD)param1, 
                            dataWR, 
                            NULL);
        break;

      case N_XIR:
      case N_XDR:
        result = BDI_DoJtag(JTAG_CMD_CLOCK, 0, 1, NULL, NULL);
        result = BDI_DoJtag((BYTE)((command == N_XIR) ? JTAG_CMD_IRSCAN : JTAG_CMD_DRSCAN),
                            0,
                            (WORD)param1, 
                            dataWR, 
                            dataRD);
        break;

      case N_HELP:
        JtagHelp();
        break;

      case N_EXIT:
        result = BDI_DoJtag(JTAG_CMD_CLOSE, 0, 0, NULL, NULL);
        modeJTAG = FALSE;
        break;

      default:
        WriteLine("Sorry, command not jet implemented!");
        break;

    } /* switch */

    /* display error */
    if (result != BDI_OKAY) {
      DisplayError(result, 0);
    } /* if */

    /* display data */
    if ((command == N_XIR) || (command == N_XDR) || (command == N_SCAN)) {
      index = (command == N_SCAN) ? param2 : param1;
      index = (index + 7) / 8;
      putPtr = szData;
      while (index > 0) {
        index--;
        *putPtr++ = BinToHex[(dataRD[index] >> 4) & 0x0F];
        *putPtr++ = BinToHex[dataRD[index] & 0x0F];
      } /* while */
      *putPtr = 0;
      WriteLine(szData);
    } /* if */

  } /* if */
  else {
    WriteLine("# invalid command in JTAG mode");
    WriteLine("# enter exit to terminate JTAG mode");
  } /* else */

} /* ProcessJTAG */


/*************************************************************************
|  MAIN
|*************************************************************************/

void main( int argc, char *argv[ ])
{
  char  rxLine[128];
  char  connectInfo[80];
  int   result;

  /* connect to BDI */
  printf("Connecting to BDI, please wait...\r");
  if (argc == 3) {
    sprintf(connectInfo, "%s %s", argv[1], argv[2]);
  } /* if */
  else if (argc == 4) {
    sprintf(connectInfo, "%s %s %s", argv[1], argv[2], argv[3]);
  } /* if */
  else  {
    sprintf(connectInfo, "NETWORK BDI3000");
  } /* if */
  result = BDI_Connect(connectInfo);
  if (result != 0) {
    printf("Connecting to BDI failed with error %i\n", result);
    printf("press enter to continue...");
    gets(rxLine);
    exit(0);
  } /* if */

  WriteLine("BDI Monitor for ARM V1.00");
  WriteLine("=========================");
  LineFeeds(1);
  WriteLine("enter HELP for command list");
  LineFeeds(1);

  terminate = FALSE;
  do {
    printf((modeJTAG) ? "JTAG>" : "BDI>");
    if (ReadLine(sizeof rxLine - 1, rxLine)) {
      if (modeJTAG) ProcessJTAG(rxLine);
      else          ProcessCommand(rxLine);
    } /* if */
  } while (!terminate);

  result = BDI_Disconnect();
} /* main */